import os
import re
import subprocess
import glob
from copy import deepcopy

import sys
if sys.version_info.major >= 3:
    from functools import reduce
import astropy.io.fits as fits
from astropy import wcs
import numpy as np
import scipy.ndimage as ndimage
import scipy.stats
import random as rd


#different imports depending on if python2.7 or python3
import sys
from copy import copy

if sys.version_info < (3,0):
    #python 2.7 behavior
    import ConfigParser
    from pyklip.instruments.Instrument import Data
    from pyklip.instruments.utils.nair import nMathar
else:
    import configparser as ConfigParser
    from pyklip.instruments.Instrument import Data
    from pyklip.instruments.utils.nair import nMathar

from pyklip.instruments.P1640_support import P1640spots
from pyklip.instruments.P1640_support import P1640utils
# from pyklip.instruments.P1640_support import P1640_cube_checker

from scipy.interpolate import interp1d
from pyklip.parallelized import high_pass_filter_imgs

class P1640Data(Data):
    """
    Note: update object string when output is decided
    A sequence of P1640 Data. Each P1640Data object has the following fields and functions

    Args:
        filepaths: list of filepaths to occulted files
        skipslices: a list of datacube slices to skip (supply index numbers e.g. [0,1,2,3])
        corefilepaths: a list of filepaths to core (i.e. unocculted) files, for contrast calc
        spot_directory: (None) path to the directory where the spot positions are stored. Defaults to P1640.ini val
        verbose: [False] if True, print more stuff
    Attributes:
        input: Array of shape (N,y,x) for N images of shape (y,x)
        centers: Array of shape (N,2) for N centers in the format [x_cent, y_cent]
        filenums: Array of size N for the numerical index to map data to file that was passed in
        filenames: Array of size N for the actual filepath of the file that corresponds to the data
        PAs: Array of N for the parallactic angle rotation of the target (used for ADI) [in degrees]
        wvs: Array of N wavelengths of the images (used for SDI) [in microns]. For polarization data, defaults to "None"
        wcs: Array of N wcs astormetry headers for each image.
        IWA: a floating point scalar (not array). Specifies to inner working angle in pixels
        output: Array of shape (b, len(files), len(uniq_wvs), y, x) where b is the number of different KL basis cutoffs
        spot_flux: Array of N of average satellite spot flux for each frame
        contrast_scaling: Flux calibration factors (multiply by image to "calibrate" flux)
        flux_units: units of output data [DN, contrast]
        prihdrs: not used for P1640, set to None
        exthdrs: Array of N P1640 headers (these are written by the P1640 cube extraction pipeline)

    Methods:
        readdata(): reread in the data
        savedata(): save a specified data in the P1640 datacube format (in the 1st extension header)
        calibrate_output(): calibrates flux of self.output
    """
    ##########################
    ###Class Initilization ###
    ##########################
    #some static variables to define the P1640 instrument
    centralwave = {}  # in microns
    fpm_diam = {}  # in pixels
    flux_zeropt = {}
    spot_ratio = {} #w.r.t. central star
    lenslet_scale = 1.0 # arcseconds per pixel (pixel scale)
    ifs_rotation = 0.0  # degrees CCW from +x axis to zenith
    nchannels_all = None
    nchannels_used = None
    observatory_latitude = 0.0

    ## read in P1640 configuration file and set these static variables
    package_directory = os.path.dirname(os.path.abspath(__file__))
    configfile = package_directory + "/" + "P1640.ini"
    config = ConfigParser.ConfigParser()
    try:
        config.read(configfile)
        #get pixel scale
        lenslet_scale = float(config.get("instrument", "ifs_lenslet_scale"))  # arcsecond/pix
        #get IFS rotation
        ifs_rotation = float(config.get("instrument", "ifs_rotation")) #degrees
        #get some information specific to each band
        bands = ['H']
        for band in bands:
            centralwave[band] = float(config.get("instrument", "cen_wave_{0}".format(band)))
            fpm_diam[band] = float(config.get("instrument", "fpm_diam")) / lenslet_scale  # pixels
        flux_zeropt = float(config.get("instrument", "zero_pt_flux"))
        observatory_latitude = float(config.get("observatory", "observatory_lat"))
        nchannels_all = int(config.get("instrument", "cube_channels"))
    except ConfigParser.Error as e:
        print("Error reading P1640 configuration file: {0}".format(e.message))
        raise e


    ####################
    ### Constructors ###
    ####################
    def __init__(self, filepaths=None, skipslices=None, corefilepaths=None, spot_directory=None, highpass=True,
                 numthreads=-1, PSF_cube=None, verbose=False):
        """
        Initialization code for P1640Data

        Note:
            Information on arguments are available in the class docstring
        """
        super(P1640Data, self).__init__()
        self._output = None

        self.flipx = True ### TODO: Is this true?

        # P1640 stuff
        self.corefilenames = corefilepaths
        self.spot_directory = spot_directory
        self.spot_flux = None # Currently not implemented, may be in future
        self.scale_factors = None # scaling between wavelength slices
        self.spot_positions = None # Ncube x Nchan x 4 x 2 spot positions [row, col]
        #self.spot_scaling = None # scaling factor between wavelengths

        self.channels_all = list(range(P1640Data.nchannels_all))
        try:
            self.channels_used = [i for i in self.channels_all if i not in skipslices]
        except:
            self.channels_used = self.channels_all[:]
        self.nchannels_used = len(self.channels_used)

        if filepaths is None:
            # general stuff
            self._input = None
            self._centers = None
            self._filenums = None
            self._filenames = None
            self._PAs = None
            self._wvs = None
            self._wcs = None
            self._IWA = None
            self.corefilenames = corefilepaths
            self.spot_flux = None # Currently not implemented, may be in future
            self.scale_factors = None # scaling factor between wavelengths
            self.contrast_scaling = None # Currently not implemented, may be in future
            self.prihdrs = None # Not used by P1640
            self.exthdrs = None # for P1640 this is the prihdrs; exthdrs used for compatibility with P1640 class
            self.flux_units = None # Currently not implemented, may be in future
        else:
            self.readdata(filepaths, skipslices=skipslices, highpass=highpass, numthreads=numthreads, PSF_cube=PSF_cube)


    ################################
    ### Instance Required Fields ###
    ################################
    @property
    def input(self):
        return self._input
    @input.setter
    def input(self, newval):
        self._input = newval

    @property
    def centers(self):
        return self._centers
    @centers.setter
    def centers(self, newval):
        self._centers = newval
        
    @property
    def filenums(self):
        return self._filenums
    @filenums.setter
    def filenums(self, newval):
        self._filenums = newval

    @property
    def filenames(self):
        return self._filenames
    @filenames.setter
    def filenames(self, newval):
        self._filenames = newval

    @property
    def configfile(self):
        return self._configfile
    @configfile.setter
    def configfile(self, newval):
        self._configfile = newval

    @property
    def PAs(self):
        return self._PAs
    @PAs.setter
    def PAs(self, newval):
        self._PAs = newval

    @property
    def wvs(self):
        return self._wvs
    @wvs.setter
    def wvs(self, newval):
        self._wvs = newval

    @property
    def wcs(self):
        return self._wcs
    @wcs.setter
    def wcs(self, newval):
        self._wcs = newval

    @property
    def IWA(self):
        return self._IWA
    @IWA.setter
    def IWA(self, newval):
        self._IWA = newval

    @property
    def output(self):
        return self._output
    @output.setter
    def output(self, newval):
        self._output = newval

    # P1640 stuff
    @property
    def corefilenames(self):
        return self._corefilenames
    @corefilenames.setter
    def corefilenames(self, newval):
        self._corefilenames = newval

    @property
    def scale_factors(self):
        return self._scale_factors
    @scale_factors.setter
    def scale_factors(self, newval):
        self._scale_factors = newval

    @property
    def spot_locations(self):
        return self._spot_locations
    @spot_locations.setter
    def spot_locations(self, newval):
        self._spot_locations = newval

    @property
    def spot_directory(self):
        return self._spot_directory
    @spot_directory.setter
    def spot_directory(self, newval):
        #if newval is None:
        #    newval = P1640Data.config("spots","spot_file_path")
        self._spot_directory = newval
        print("Spot file directory set to {0}".format(self.spot_directory))

    ###############
    ### Methods ###
    ###############
    def get_scaling_and_centering(self, filepaths, spot_directory=None, skipslices=None):
        """
        Method to get star position and wavelength scaling factors
        Args:
            filepaths: a list of filepaths to datacubes
            spot_path: (None) directory where the spot files are stored. Defaults to P1640.ini value
            skipslices: a list of wavelength slices to skip for each datacube (supply index numbers e.g. [0,1,2,3])
        Returns:
            Nothing. Sets centers and scale_factors fields of the P1640Data object
        """
        if isinstance(filepaths, str):
            filepaths = [filepaths]

        scale_factors = []
        centers = []
        spot_photometry = []

        for filepath in filepaths:
            hdulist = fits.open(filepath)
            cube = hdulist[0].data

            try:
                if spot_directory is not None:
                    spot_filedir = spot_directory
                    print(spot_filedir)
                else: # use the default set in P1640.ini
                    spot_filedir = P1640Data.config.get("spots","spot_filepath")
                spot_filebasename = os.path.splitext(os.path.basename(filepath))[0]
                spot_fullpath = os.path.join(spot_filedir, spot_filebasename)
                spot_filepaths = glob.glob(spot_fullpath+'*')

                # check if all the spot files exist, if so, read them in
                exist = np.all([os.path.isfile(f) for f in spot_filepaths])
                assert(exist is not False)
                # print("Reading spots from files: {0}".format(os.path.commonprefix(spot_filepaths)))
                spot_locations = np.array([np.genfromtxt(f, delimiter=',') 
                                           for f in spot_filepaths])
                # return spot_locations
            except AssertionError:
                # if they haven't already been written to file, calculate them
                spot_locations = P1640spots.get_single_cube_spot_positions(cube)
                # write them to disk so they don't have to be recalculated
                write_p1640_spots_to_file(P1640Data.config, filepath, spot_locations)

            cube_spot_fluxes = P1640spots.get_single_cube_spot_photometry(cube, spot_locations)
            cube_scale, cube_center = P1640spots.get_scaling_and_centering_from_spots(spot_locations)

            hdulist.close()

            scale_factors.append(cube_scale)
            centers.append(cube_center)
            spot_photometry.append(cube_spot_fluxes)
        return centers
        #self.scale_factors = scale_factors
        #self.centers = centers
        #self.spot_flux = cube_spot_fluxes


    def readdata(self, filepaths, skipslices=None, corefilepaths=None, highpass=True, numthreads=-1, PSF_cube=None, verbose=False):
        """
        Method to open and read a list of P1640 data. Handles everything that can be done by
        reading directly from the P1640 header or cubes, no calculations. Scaling and Centering handled elsewhere

        Args:
            filepaths: a list of filepaths
            skipslices: a list of wavelenegth slices to skip for each datacube (supply index numbers e.g. [0,1,2,3])
            highpass: if True, run a Gaussian high pass filter (default size is sigma=imgsize/10)
                      can also be a number specifying FWHM of box in pixel units
            numthreads: Number of threads to be used. Default -1 sequential sat spot flux calc.
                        If None, numthreads = mp.cpu_count().
            PSF_cube: 3D array (nl,ny,nx) with the PSF cube to be used in the flux calculation.
        Returns:
            Technically none. It saves things to fields of the P1640Data object. See object doc string
        """
        #check to see if user just inputted a single filename string
        if isinstance(filepaths, str):
            filepaths = [filepaths]

        #make some lists for quick appending
        data = []
        filenums = []
        filenames = []
        rot_angles = []
        wvs = []
        centers = []
        spot_scalings = []
        spot_locations = []
        wcs_hdrs = []
        spot_fluxes = []
        prihdrs = []
        exthdrs = []

        #extract data from each file
        for index, filepath in enumerate(filepaths):
            cube, center, spot_scaling_single_cube, spot_locations_single_cube, pa, wv, \
            astr_hdrs, filt_band, fpm_band, ppm_band, spot_flux, prihdr, exthdr = \
                _p1640_process_file(filepath, spot_directory=self.spot_directory, skipslices=skipslices,
                                    highpass=highpass, numthreads=numthreads, verbose=verbose)

            data.append(cube)
            centers.append(center)
            spot_scalings.append(spot_scaling_single_cube)
            spot_locations.append(spot_locations_single_cube)
            spot_fluxes.append(spot_flux)
            rot_angles.append(pa)
            wvs.append(wv)
            filenums.append(np.ones(pa.shape[0]) * index)
            wcs_hdrs.append(astr_hdrs)
            prihdrs.append(prihdr)
            exthdrs.append(exthdr)
            filenames.append([filepath for i in range(pa.shape[0])])
        if verbose == False:
            print("{N} files processed from {p} (verbose = False).".format(N=len(filepaths),
                                                                           p=os.path.commonprefix(filepaths)))


        #convert everything into numpy arrays
        #reshape arrays so that we collapse all the files together (i.e. don't care about distinguishing files)
        data = np.array(data)
        data = P1640utils.set_zeros_to_nan(data)
        dims = data.shape
        data = data.reshape([dims[0] * dims[1], dims[2], dims[3]])
        filenums = np.array(filenums).reshape([dims[0] * dims[1]])
        filenames = np.array(filenames).reshape([dims[0] * dims[1]])
        rot_angles = -(np.array(rot_angles).reshape([dims[0] * dims[1]])) + (90 - self.ifs_rotation) # want North Up
        wvs = np.array(wvs).reshape([dims[0] * dims[1]])
        wcs_hdrs = np.array(wcs_hdrs).reshape([dims[0] * dims[1]])
        centers = np.array(centers).reshape([dims[0] * dims[1], 2])
        spot_fluxes = np.array(spot_fluxes).reshape([dims[0] * dims[1]])
        spot_locations = np.rollaxis(np.array(spot_locations), -2, 1).reshape([dims[0]*dims[1], 4, 2])[..., ::-1]
        # for the spot scalings factors, take the mean for each wavelength
        spot_scalings = np.array(spot_scalings)#.reshape([dims[0]*dims[1]])
        # Not used by P1640
        '''
        #only do the wavelength solution and center recalculation if it isn't broadband imaging
        if np.size(np.unique(wvs)) > 1:
            # recalculate wavelengths from satellite spots
            wvs = rescale_wvs(exthdrs, wvs, skipslices=skipslices)
            # recaclulate centers from satellite spots and new wavelength solution
            wvs_bycube = wvs.reshape([dims[0], dims[1]])
            centers_bycube = centers.reshape([dims[0], dims[1], 2])
            for i, cubewvs in enumerate(wvs_bycube):
                try:
                    centers_bycube[i] = calc_center(prihdrs[i], exthdrs[i], cubewvs, skipslices=skipslices)
                except KeyError:
                    print("Unable to recenter the data using a least squraes fit due to not enough header info for file "
                          "{0}".format(filenames[i*dims[1]]))
        '''
        
        #set these as the fields for the P1640Data object
        self._input = data
        self._centers = centers
        self._filenums = filenums
        self._filenames = filenames
        self._corefilenames = corefilepaths
        self._PAs = rot_angles
        # use the spot scalings instead of the wavelength solution; that comes separately anyway
        #self._wvs =  wvs
        self._wvs = np.tile(np.mean(spot_scalings, axis=0), spot_scalings.shape[0])
        self._wcs = wcs_hdrs 
        self._IWA = P1640Data.fpm_diam[fpm_band]/2.0
        self.spot_flux = spot_fluxes
        self.scale_factors = spot_scalings
        self.spot_locations = spot_locations
        self.contrast_scaling = None #P1640Data.spot_ratio[ppm_band]/np.mean(spot_fluxes)
        self.flux_units = "DN"
        self.prihdrs = prihdrs
        self.exthdrs = exthdrs

    def savedata(self, filepath, data, klipparams = None, filetype = 'PSF Subtracted Spectral Cube',
                 zaxis = None, center=None, astr_hdr=None, fakePlparams = None, more_keywords=None,
                 user_prihdr = None, user_exthdr = None, extra_exthdr_keywords = None,
                 extra_prihdr_keywords = None ):
        """
        Save data in a fits file in a GPI-like fashion. Aka, data and header are in the extension HDU.
        For now, the Primary HDU contains the KLIP parameters, scaling, and centering. This may later change 
        storing the data in the Primary HDU, all the headers from the input files in the extension.

        Note: In principle, the function only works inside klip_dataset(). In order to use it outside of klip_dataset,
            you need to define the follwing attributes:
                dataset.output_wcs = np.array([w.deepcopy() if w is not None else None for w in dataset.wcs])
                dataset.output_centers = dataset.centers

        Inputs:
            filepath: path to file to output
            data: 2D or 3D data to save
            klipparams: a string of klip parameters
            filetype: filetype of the object (e.g. "KL Mode Cube", "PSF Subtracted Spectral Cube")
            zaxis: a list of values for the zaxis of the datacube (for KL mode cubes currently)
            astr_hdr: wcs astrometry header
            center: center of the image to be saved in the header as the keywords PSFCENTX and PSFCENTY in pixels.
                The first pixel has coordinates (0,0)
            fakePlparams: fake planet params

        """
        hdulist = fits.HDUList()
        hdulist.append(fits.PrimaryHDU(header=None, data=data))
        # save all the files we used in the reduction
        # we'll assume you used all the input files
        # remove duplicates from list
        filenames = np.unique(self.filenames)
        nfiles = np.size(filenames)
        hdulist[0].header["DRPNFILE"] = (nfiles, "Num raw files used in pyKLIP")
        for i, thispath in enumerate(filenames):
            thispath = thispath.replace("\\", '/')
            splited = thispath.split("/")
            fname = splited[-1]
            matches = re.search('20[0-9]{2}-[0-9]{2}-[0-9]{2}_[0-9]{3}', fname)
            filename = matches.group(0)
            hdulist[0].header["FILE_{0}".format(i)] = filename + '.fits'

        # write out psf subtraction parameters
        # get pyKLIP revision number
        pykliproot = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        # the universal_newline argument is just so python3 returns a string instead of bytes
        # this will probably come to bite me later
        try:
            pyklipver = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD'], cwd=pykliproot, universal_newlines=True).strip()
        except:
            pyklipver = "unknown"
        hdulist[0].header['PSFSUB'] = ("pyKLIP", "PSF Subtraction Algo")
        hdulist[0].header.add_history("Reduced with pyKLIP using commit {0}".format(pyklipver))
        if self.creator is None:
            hdulist[0].header['CREATOR'] = "pyKLIP-{0}".format(pyklipver)
        else:
            hdulist[0].header['CREATOR'] = self.creator
            hdulist[0].header.add_history("Reduced by {0}".format(self.creator))

        # store commit number for pyklip
        hdulist[0].header['pyklipv'] = (pyklipver, "pyKLIP version that was used")

        if klipparams is not None:
            hdulist[0].header['PSFPARAM'] = (klipparams, "KLIP parameters")
            hdulist[0].header.add_history("pyKLIP reduction with parameters {0}".format(klipparams))

        if fakePlparams is not None:
            hdulist[0].header['FAKPLPAR'] = (fakePlparams, "Fake planet parameters")
            hdulist[0].header.add_history("pyKLIP reduction with fake planet injection parameters {0}".format(fakePlparams))
        # store file type
        if filetype is not None:
            hdulist[0].header['FILETYPE'] = (filetype, "P1640 File type")

        if self.channels_used is not None:
            hdulist[0].header['channels'] = ','.join(['{0}'.format(i) for i in self.channels_used])

        #write flux units/conversion
        hdulist[0].header['FUNIT'] = (self.flux_units, "Flux units of data")
        if self.flux_units.upper() == 'CONTRAST':
            hdulist[0].header['DN2CON'] = (self.contrast_scaling, "Contrast/DN")
            hdulist[0].header.add_history("Converted to contrast units using {0} Contrast/DN".format(self.contrast_scaling))
            core_header = fits.Header()
            for i,cf in enumerate(self._corefilenames):
                core_header['Core_{0:02d}'.format(i)] = cf
            core_header['Method'] = 'median'
            core_hdu = fits.ImageHDU(self._core_psf, name='Core')
            core_hdu.header = core_header
            hdulist.append(fits.ImageHDU(self._core_psf))

        # write z axis units if necessary
        if zaxis is not None:
            #Writing a KL mode Cube
            if "KL Mode" in filetype:
                hdulist[0].header['CTYPE3'] = 'KLMODES'
                #write them individually
                for i, klmode in enumerate(zaxis):
                    hdulist[0].header['KLMODE{0}'.format(i)] = (klmode, "KL Mode of slice {0}".format(i))

        #use the dataset astr hdr if none was passed in
        if astr_hdr is None:
            astr_hdr = self.output_wcs[0]
        if astr_hdr is not None:
            #update astro header
            #I don't have a better way doing this so we'll just inject all the values by hand
            astroheader = astr_hdr.to_header()
            exthdr = hdulist[0].header
            exthdr['PC1_1'] = astroheader['PC1_1']
            exthdr['PC2_2'] = astroheader['PC2_2']
            try:
                exthdr['PC1_2'] = astroheader['PC1_2']
                exthdr['PC2_1'] = astroheader['PC2_1']
            except KeyError:
                exthdr['PC1_2'] = 0.0
                exthdr['PC2_1'] = 0.0
            # remove CD values as those are confusing
            # currently not found in P1640 headers
            try:
                exthdr.remove('CD1_1')
                exthdr.remove('CD1_2')
                exthdr.remove('CD2_1')
                exthdr.remove('CD2_2')
            except:
                pass
            exthdr['CDELT1'] = 1
            exthdr['CDELT2'] = 1

        #use the dataset center if none was passed in
        if center is None:
            center = self.output_centers[0]
        if center is not None:
            hdulist[0].header.update({'PSFCENTX':center[0],'PSFCENTY':center[1]})
            hdulist[0].header.update({'CRPIX1':center[0],'CRPIX2':center[1]})
            hdulist[0].header.add_history("Image recentered to {0}".format(str(center)))
        #if scaling is None:
        #    scaling = self.scaling

        # store the wavelength solution in a TableHDU
        wvs_hdu = fits.TableHDU.from_columns([fits.Column(name='wavelength',
                                                          format='E',
                                                          array=self.wvs)],
                                             name='wavelengths')

        hdulist.append(wvs_hdu)
        # store all headers from the source files
        for i, ext in enumerate(self.exthdrs):
            hdulist.append(fits.ImageHDU(header=ext, name="Data_{0:02d}".format(i)))


        try:
            hdulist.writeto(filepath, overwrite=True)
        except TypeError:
            hdulist.writeto(filepath, clobber=True)
        hdulist.close()

    def calibrate_output(self, units="contrast"):
        """
        Calibrates the flux of the output of PSF subtracted data.

        Assumes self.output exists and has shape (b,N,y,x) for N is the number of images and b is
        number of KL modes used.

        Args:
            units: currently only support "contrast" w.r.t central star

        Returns:
            stores calibrated data in self.output
        """
        if units == "contrast":
            # assemble median core
            try:
                assert(self._corefilenames is not None)
            except AssertionError:
                "No core files = no calibration possible"
                return
            if len(self._corefilenames) == 1:
                core_files = [self._corefilenames]
            core_hdus = [fits.open(f) for f in self._corefilenames]
            core_cubes = np.array([core_hdus.data for hdu in core_hdus])
            star_psf = P1640cores.make_median_core(core_cubes)
            for hdu in core_hdus:
                hdu.close()

            self.contrast_scaling = star_flux = np.nansum(axis=0)


            self.output *= self.contrast_scaling
            self.flux_units = "contrast"
            self._corefilenames = core_files
            self._core_psf = star_psf


    def generate_psfs(self, boxrad=7, gauss=True, mirror_adjust=1.0):
        """
        Generates PSF for each frame of input data. Only works on spectral mode data.
        Args:
            boxrad: the halflength of the size of the extracted PSF (in pixels)
            # spotyx: Ncube x Nchan x 4 x 2 array of spot (row, col) positions
            gauss [False]: if True, use a gaussian PSF
            mirror_adjust [1.0]: multiply mirror diameter by this factor to adjust for Lyot stop
        Returns:
            saves PSFs to self.psfs as an array of size(N,psfy,psfx) where psfy=psfx=2*boxrad + 1
        """
        self.psfs = []

        # spots initially have shape (Ncube x Nchan x Nspots x 2)
        nframes = reduce(lambda x,y: x*y, self.spot_locations.shape[:-2])

        spot0 = self.spot_locations[..., 0,:].reshape(nframes, 2)
        spot1 = self.spot_locations[..., 1,:].reshape(nframes, 2)
        spot2 = self.spot_locations[..., 2,:].reshape(nframes, 2)
        spot3 = self.spot_locations[..., 3,:].reshape(nframes, 2)

        if gauss == True: # simulate the PSF as a gaussian
            stamp = 2*boxrad + 1
            uniqwvs = np.unique(self.wvs)
            nl = uniqwvs.size
            radial_psfs = np.zeros((nl, stamp, stamp))
            telD = self.config.getfloat('observatory','primary_diam')
            telD *= mirror_adjust # undersize by 10% for Lyot stop
            for wv, lam in enumerate(uniqwvs):
                # Calculate lam/D in pixels - first convert wavelength to [m]
                # lam[m] / D[m] is in radians -- convert to arcsec
                fwhm_arcsec = 1.22*((lam*1.0e-6)/telD) * (3600*180/np.pi)
                # convert to pixels with ifs_lenslet_scale
                fwhm = fwhm_arcsec/self.config.getfloat('instrument','ifs_lenslet_scale')
                # Gaussian standard deviation - from another routine
                sigma = fwhm/(2.*np.sqrt(2*np.log(2)))
                #centered in the array
                y,x = np.indices([stamp, stamp])
                y -= stamp // 2
                x -= stamp // 2
                radial_psfs[wv,...] = np.exp(-(x**2. + y**2.) / (2. * sigma**2))
            self.psfs = radial_psfs / np.mean(radial_psfs.sum(axis=0))
        else: # use the spots to build the psf
            for i,frame in enumerate(self.input):
                spots = [[float(spot0[i, 0]), float(spot0[i, 1])], [float(spot1[i, 0]), float(spot1[i, 1])],
                         [float(spot2[i, 0]), float(spot2[i, 1])], [float(spot3[i, 0]), float(spot3[i, 1])]]
                #now make a psf
                spotpsf = generate_psf(frame, spots, boxrad=boxrad)
                self.psfs.append(spotpsf)
            self.psfs = np.array(self.psfs)
            self.psfs = np.reshape(self.psfs, (self.psfs.shape[0]//self.nchannels_used, self.nchannels_used,
                                               self.psfs.shape[1], self.psfs.shape[2]))
            self.psfs = np.mean(self.psfs, axis=0)

    def generate_psf_cube(self, boxw=14):
        return generate_psfs(int(np.floor(boxw//2)))
    """
    def generate_psf_cube(self, boxw=14):
        \"""
        # P1640 - use unocculted files
        Generates an average PSF from all frames of input data. Only works on spectral mode data.
        Overall cube normalized to unity with norm 2.
        Currently hard coded assuming 37 spectral channels!!!

        The center of the PSF is exactly on the central pixel of the PSF.
        (If even width of the array it is the middle pixel with the highest row and column index.)
        The center pixel index is always (nx/2,nx/2) assuming integer division.

        Args:
            boxw: the width the extracted PSF (in pixels). Should be bigger than 12 because there is an interpolation
                of the background by a plane which is then subtracted to remove linear biases.

        Returns:
            A cube of shape Nchan*boxw*boxw. Each slice [k,:,:] is the PSF for a given wavelength.
        \"""

        n_frames,ny,nx = self.input.shape
        x_grid, y_grid = np.meshgrid(np.arange(ny), np.arange(nx))
        unique_wvs = np.unique(self.wvs)
        numwaves = np.size(self.channels_used)  # np.unique(self.wvs))

        psfs = np.zeros((numwaves,boxw,boxw,n_frames,4))

        # spots initially have shape (Ncube x Nchan x Nspots x 2)
        nframes = reduce(lambda x,y: x*y, self.spot_locations.shape[:-2])
        spot0 = self.spot_locations[..., 0,:].reshape(nframes, 2)
        spot1 = self.spot_locations[..., 1,:].reshape(nframes, 2)
        spot2 = self.spot_locations[..., 2,:].reshape(nframes, 2)
        spot3 = self.spot_locations[..., 3,:].reshape(nframes, 2)

        scale_factors = self.scale_factors.ravel()
        for lambda_ref_id, lambda_ref in enumerate(self.channels_used):
            for i,frame in enumerate(self.input):
                #figure out which header and which wavelength slice
                #put all the sat spot info together
                spots = [[float(spot0[i, 0]), float(spot0[i, 1])], [float(spot1[i, 0]), float(spot1[i, 1])],
                         [float(spot2[i, 0]), float(spot2[i, 1])], [float(spot3[i, 0]), float(spot3[i, 1])]]
                #mask nans
                cleaned = np.copy(frame)
                cleaned[np.where(np.isnan(cleaned))] = 0

                for loc_id, loc in enumerate(spots):
                    #grab satellite spot positions
                    spotx = loc[0]
                    spoty = loc[1]
                    xarr_spot = np.round(spotx)
                    yarr_spot = np.round(spoty)
                    stamp = cleaned[(yarr_spot-np.floor(boxw/2.0)):(yarr_spot+np.ceil(boxw/2.0)),(xarr_spot-np.floor(boxw/2.0)):(xarr_spot+np.ceil(boxw/2.0))]
                    #x_stamp = x_grid[(yarr_spot-boxw/2):(yarr_spot+boxw/2),(xarr_spot-boxw/2):(xarr_spot+boxw/2)]
                    #y_stamp = y_grid[(yarr_spot-boxw/2):(yarr_spot+boxw/2),(xarr_spot-boxw/2):(xarr_spot+boxw/2)]
                    #print(spotx,spoty)
                    #print(stamp_x+ spotx-xarr_spot,stamp_y+spoty-yarr_spot)
                    stamp_x, stamp_y = np.meshgrid(np.arange(boxw, dtype=np.float32), np.arange(boxw, dtype=np.float32))
                    dx = spotx-xarr_spot
                    dy = spoty-yarr_spot
                    #stamp_x += spotx-xarr_spot
                    #stamp_y += spoty-yarr_spot
                    #stamp_x -= spotx-xarr_spot
                    #stamp_y -= spoty-yarr_spot
                    #print(spotx-xarr_spot,spoty-yarr_spot)


                    #mask the central blob to calculate background median
                    stamp_r = np.sqrt((stamp_x-dx-boxw/2)**2+(stamp_y-dy-boxw/2)**2)
                    stamp_masked = copy(stamp)
                    stamp_x_masked = stamp_x-dx
                    stamp_y_masked = stamp_y-dy
                    stamp_center = np.where(stamp_r<4)
                    stamp_masked[stamp_center] = np.nan
                    stamp_x_masked[stamp_center] = np.nan
                    stamp_y_masked[stamp_center] = np.nan
                    background_med =  np.nanmedian(stamp_masked)
                    stamp_masked -= background_med
                    #Solve 2d linear fit to remove background
                    xx = np.nansum(stamp_x_masked**2)
                    yy = np.nansum(stamp_y_masked**2)
                    xy = np.nansum(stamp_y_masked*stamp_x_masked)
                    xz = np.nansum(stamp_masked*stamp_x_masked)
                    yz = np.nansum(stamp_y_masked*stamp_masked)
                    #Cramer's rule
                    a = (xz*yy-yz*xy)/(xx*yy-xy*xy)
                    b = (xx*yz-xy*xz)/(xx*yy-xy*xy)
                    stamp -= a*(stamp_x-dx)+b*(stamp_y-dy) + background_med
                    #stamp -= background_med

                    #rescale to take into account wavelength widening
                    if 1:
                        stamp_r = np.sqrt((stamp_x-dx-boxw/2)**2+(stamp_y-dy-boxw/2)**2)
                        stamp_th = np.arctan2(stamp_x-dx-boxw/2,stamp_y-dy-boxw/2)
                        stamp_r /= scale_factors[i]  # lambda_ref/lambda_curr
                        stamp_x = stamp_r*np.cos(stamp_th)+boxw/2
                        stamp_y = stamp_r*np.sin(stamp_th)+boxw/2
                        #print(stamp_x,stamp_y)

                    stamp = ndimage.map_coordinates(stamp, [stamp_y+dx, stamp_x+dy])
                    #print(stamp)
                    psfs[lambda_ref_id,:,:,i,loc_id] = stamp


        #PSF_cube = np.mean(psfs[:,:,:,:,0],axis=(3))
        PSF_cube = np.mean(psfs,axis=(3,4))

        #Build the spectrum of the sat spots
        # Number of cubes in dataset
        N_cubes = int(self.input.shape[0])/int(numwaves)
        N_chans = np.size(self.channels_all)
        all_sat_spot_spec = np.zeros((N_chans, N_cubes))
        for k in range(N_cubes):
            all_sat_spot_spec[:,k] = self.spot_flux[N_chans*k: N_chans*(k+1)]
        sat_spot_spec = np.nanmean(all_sat_spot_spec,axis=1)

        PSF_cube /= np.sqrt(np.nansum(PSF_cube**2))
        for l in range(numwaves):
            #PSF_cube[l,:,:] -= np.nanmedian(PSF_cube[l,:,:][stamp_center])
            PSF_cube[l,:,:] *= sat_spot_spec[l]/np.nanmax(PSF_cube[l,:,:])
            PSF_cube[l,:,:][np.where(abs(PSF_cube[l,:,:])/np.nanmax(abs(PSF_cube[l,:,:]))<0.05)] = 0.0

        if 0:
            import matplotlib.pyplot as plt
            plt.figure(1)
            plt.plot(sat_spot_spec,'or')
            plt.plot(np.nanmax(PSF_cube,axis=(1,2)),"--b")
            plt.show()
        if 0: # for debugging purposes
            import matplotlib.pyplot as plt
            plt.figure(1)
            plt.imshow(PSF_cube[0,:,:],interpolation = 'nearest')
            plt.figure(2)
            plt.imshow(PSF_cube[36,:,:],interpolation = 'nearest')
            plt.show()

        self.psfs = PSF_cube
    """
    def get_radial_psf(self,save = None):
        """
        Return a pure radial PSF by averaging the original psf. The new PSF is invariant by rotation.
        A call to generate_psf_cube() is required prior to calling this function.
        The center pixel index is always (nx/2,nx/2) assuming integer division.

        Args:
            save: Optionally automatically save the radial psf cube as a fits file with filename:
                    save+"-original_radial_PSF_cube.fits"

        Returns:
            rad_psf_cube: a (37,nx,nx) cube with the radial psf.

        """
        if np.size(np.shape(self.psfs)) == 3 and np.shape(self.psfs)[0] == 37:
            nl,ny,nx = self.psfs.shape
            # We should have nx = ny

            sat_spot_spec = np.nanmax(self.psfs,axis=(1,2))

            k_hd=4 # should be even
            nx_hd = k_hd*(nx-1) + 1
            hd_psf = np.zeros((nl,nx_hd,nx_hd))

            rad_psf_cube = np.zeros((nl,nx,nx))
            #current_slice = np.zeros((nx,nx))

            stamp_x, stamp_y = np.meshgrid(np.arange(nx, dtype=np.float32), np.arange(nx, dtype=np.float32))
            stamp_r = np.sqrt((stamp_x - nx/2)**2+(stamp_y - nx/2)**2)
            stamp_x_hd, stamp_y_hd = np.meshgrid(np.arange(nx_hd, dtype=np.float32)/(nx_hd-1)*(nx-1), np.arange(nx_hd, dtype=np.float32)/(nx_hd-1)*(nx-1))
            for l in range(nl):
                hd_psf[l,:,:] = ndimage.map_coordinates(self.psfs[l,:,:], [stamp_y_hd, stamp_x_hd])
                #hd_psf[l,nx/2*k_hd,nx/2*k_hd] = 0. # center
            stamp_r_hd = np.sqrt((stamp_x_hd-stamp_x_hd[nx/2*k_hd,nx/2*k_hd])**2+(stamp_y_hd-stamp_y_hd[nx/2*k_hd,nx/2*k_hd])**2)

            dr = 1.0/k_hd
            Dr = 2.0/k_hd
            r_samp = np.arange(0,np.max(stamp_r_hd)+dr,dr)

            radial_val = np.zeros((nl,np.size(r_samp)))

            for r_id, r_it in enumerate(r_samp):
                selec_pix = np.where( ((r_it-Dr/2.0) < stamp_r_hd) * (stamp_r_hd < (r_it+Dr/2.0)) )
                selec_y, selec_x = selec_pix
                radial_val[:,r_id] = np.nanmean(hd_psf[:,selec_y, selec_x],1)

            for l_id in np.arange(nl):
                f = interp1d(r_samp, radial_val[l_id,:], kind='cubic',bounds_error=False, fill_value=np.nan)
                rad_psf_cube[l_id,:,:] = f(stamp_r.reshape(nx*nx)).reshape(nx,nx)
                rad_psf_cube[l_id,:,:] *= sat_spot_spec[l_id]/np.nanmax(rad_psf_cube[l_id,:,:])

                if 0:
                    import matplotlib.pyplot as plt
                    print(rad_psf_cube[l_id,0,0])
                    plt.figure(1)
                    plt.imshow(rad_psf_cube[l_id,:,:],interpolation = 'nearest')
                    plt.figure(2)
                    plt.plot(np.nanmax(self.psfs,axis=(1,2)))
                    plt.show()



            if save is not None:
                self.savedata(save+"-original_radial_PSF_cube.fits", rad_psf_cube)

            return rad_psf_cube

        else:
            print("Wrong size of the PSFs stored in p1640 dataset structure when calling get_radial_psf. Return 0")
            return 0

######################
## Static Functions ##
######################
def get_p1640_spot_filepaths(config, data_filepath):
    """
    Look to see if the spot positions have already been written to file.
    If the files exist, return their paths; otherwise return None.
    Input:
        config: a ConfigParser object with the file path information
        data_filepath: the name of the P1640 data file whose spots you want to locate
    Returns:
        filepath: a path to the files (None if they don't exist)
    """
    try:
        spot_filedir = config.get("spots","spot_file_path")
        spot_filepostfix = config.get("spots","spot_file_postfix")
        spot_fileext = config.get("spots", "spot_file_ext")
    except ConfigParser.Error as e:
        print("Spot file path not found in P1640 config file: {0}".format(e.message))
        raise e

    spot_filebasename = os.path.splitext(os.path.basename(data_filepath))[0] + spot_filepostfix
    spot_fullpath = os.path.join(spot_filedir, spot_filebasename)
    spot_filepaths = [spot_fullpath + "{0}".format(i)+spot_fileext for i in range(4)]
    return spot_filepaths

def write_p1640_spots_to_file(config, data_filepath, spot_positions,
                              spot_directory=None, overwrite=True):
    """
    EDIT: NOW THIS IS JUST A WRAPPER FOR THE P1640spots.write_spots_to_file() METHOD

    Write the spot (row, col) positions to 4 files (1 per spot) in the directory specified
    in the config file.
    Input:
        config: a ConfigParser object with the file path information
        data_filepath: source file, spot files will have same prefix
        spot_positions: a Nspot x Nchan x 2 array of (row, col) spot positions
        spot_directory: (None) Directory to save the spot files
        overwrite: True -> overwrite (default), False -> don't overwrite existing files
    Output:
        None
    """
    if spot_directory is None:
        spot_directory = config.get("spots","spot_file_path")
    spot_filepostfix = config.get("spots","spot_file_postfix")
    spot_fileext = config.get("spots", "spot_file_ext")
    P1640spots.write_spots_to_file(data_filepath, spot_positions,
                                   spot_directory, overwrite,
                                   spot_filepostfix, spot_fileext)
    return

def _p1640_process_file(filepath, spot_directory=None, skipslices=None, highpass=True,
                        numthreads=-1, psfs_func_list=None, verbose=False):
    """
    Method to open and parse a P1640 file

    Args:
        filepath: the file to open
        spot_directory: path to folder were spot positions are stored (defaults to P1640.ini)
        skipslices: a list of datacube slices to skip (supply index numbers e.g. [0,1,2,3])
        highpass: if True, run a Gaussian high pass filter (default size is sigma = imgsize/10)
                  can also be a number specifying FWHM of box in pixel units
        numthreads: : Number of threads to be used. Default -1 sequential sat spot flux calc.
                        If None, numthreads = mp.cpu_count().
        PSF_cube: 3D array (nl,ny,nx) with the PSF cube to be used in the flux calculation.
        verbose [False]: if True, print more stuff
    Returns: (using z as size of 3rd dimension, z=32 for spec including all wavelengths)
        cube: 3D data cube from the file. Shape is (z,281,281)
        center: array of shape (z,2) giving each datacube slice a [xcenter,ycenter] in that order
        parang: array of z of the parallactic angle of the target (same value just repeated z times b/c SDI)
        wvs: array of z of the wavelength of each datacube slice. (For pol mode, wvs = [None])
        astr_hdrs: array of z of the WCS header for each datacube slice
        filt_band: the band (Y, J, H, K1, K2) used in the IFS Filter (string) NOT USED
        fpm_band: which coronagrpah was used (string) NOT USED
        ppm_band: which apodizer was used (string) NOT USED
        spot_fluxes: array of z containing average satellite spot fluxes for each image NOT USED
        spot_positions: array of (row, col) spot positions for each slice
        prihdr: primary header of the FITS file NOT USED
        exthdr: 1st extention header of the FITS file
    """
    if verbose == True:
        print("Reading File: {0}".format(filepath))
    hdulist = fits.open(filepath)
    try:
        #grab the data and headers
        cube = hdulist[0].data
        exthdr = hdulist[0].header
        prihdr = None #hdulist[0].header

        #get some instrument configuration from the primary header
        # GPI relics, kept for compatibility
        filt_band ='H' # None #prihdr['IFSFILT'].split('_')[1]
        fpm_band = 'H' #None #prihdr['OCCULTER'].split('_')[1]
        ppm_band = 'H' #None #prihdr['APODIZER'].split('_')[1] #to determine sat spot ratios

        #grab the astro header
        w = wcs.WCS(header=exthdr, naxis=[1,2])
        #turns out WCS data can be wrong. Let's recalculate it using avparang
        parang = 0 # for P1640 #exthdr['AVPARANG']
        # WCS information not saved by P1640
        vert_angle = -(360-parang) + P1640Data.ifs_rotation - 90
        vert_angle = np.radians(vert_angle)
        pc = np.array([[np.cos(vert_angle), np.sin(vert_angle)],[-np.sin(vert_angle), np.cos(vert_angle)]])
        cdmatrix = pc * P1640Data.lenslet_scale /3600.
        # P1640 workaround
        w.wcs.cd = cdmatrix
        """ old version
        w.wcs.cd[0,0] = cdmatrix[0,0]
        w.wcs.cd[0,1] = cdmatrix[0,1]
        w.wcs.cd[1,0] = cdmatrix[1,0]
        w.wcs.cd[1,1] = cdmatrix[1,1]
        """

        channels = exthdr['NAXIS3']
        wvs = P1640spots.P1640params.wlsol #get wavelength solution
        # calculate centers from satellite spots
        # first, check if spot positions have been stored on disk
        # build the path
        spot_locations = []
        try:
            if spot_directory is not None:
                spot_filedir = spot_directory
                if verbose == True:
                    print("Spot directory: {s}".format(s=spot_filedir))
            else: # use the default set in P1640.ini
                spot_filedir = P1640Data.config.get("spots","spot_file_path")
                #spot_filepaths = get_p1640_spot_filepaths(P1640Data.config, filepath)
            spot_filebasename = os.path.splitext(os.path.basename(filepath))[0]
            spot_fullpath = os.path.join(spot_filedir, spot_filebasename)
            spot_filepaths = sorted(glob.glob(spot_fullpath+'*'))

            # check if all the spot files exist, if so, read them in
            exist = np.all([os.path.isfile(f) for f in spot_filepaths])
            assert(exist is not False)
            if verbose == True:
                print("Reading spots from files: {0}".format(os.path.commonprefix(spot_filepaths)))
            spot_locations = np.array([np.genfromtxt(f, delimiter=',') 
                                       for f in spot_filepaths])
        except AssertionError:
            # if they haven't already been written to file, calculate them
            spot_locations = P1640spots.get_single_cube_spot_positions(cube)
            # write them to disk so they don't have to be recalculated
            write_p1640_spots_to_file(P1640Data.config, filepath, spot_locations)
        spot_fluxes = P1640spots.get_single_cube_spot_photometry(cube, spot_locations)
        scale_factors, center = P1640spots.get_scaling_and_centering_from_spots(spot_locations,
                                                                                mean_scaling=False)

        #parang = np.repeat(exthdr['AVPARANG'], channels) #populate PA for each wavelength slice (the same)
        parang = np.repeat(0, channels) #populate PA for each wavelength slice (the same)
        astr_hdrs = [w.deepcopy() for i in range(channels)] #repeat astrom header for each wavelength slice
    finally:
        hdulist.close()

    scale_factors = np.squeeze(scale_factors).mean(axis=-2)
    spot_fluxes = np.mean(spot_fluxes, axis=-2)
    
    #remove undesirable slices of the datacube if necessary
    if skipslices is not None:
        cube = np.delete(cube, skipslices, axis=0)
        center = np.delete(center, skipslices, axis=0)
        parang = np.delete(parang, skipslices)
        wvs = np.delete(wvs, skipslices)
        astr_hdrs = np.delete(astr_hdrs, skipslices)
        spot_fluxes = np.delete(spot_fluxes, skipslices, axis=0)
        spot_locations = np.delete(spot_locations, skipslices, axis=1)
        scale_factors = np.delete(scale_factors, skipslices, axis=0)
    highpassed = False
    if isinstance(highpass, bool):
        if highpass:
            cube = high_pass_filter_imgs(cube)
            highpassed = True
    else:
        # should be a number
        if isinstance(highpass, (float, int)):
            fourier_sigma_size = (cube.shape[1]/(highpass)) / (2*np.sqrt(2*np.log(2)))
            cube = high_pass_filter_imgs(cube, filtersize=fourier_sigma_size)
            highpassed = True

    # pyklip centers need to be [x,y] instead of (row, col)
    #print(center.shape)
    center = np.fliplr(center) # [0] because of the way P1640spots works
    return cube, center, scale_factors, spot_locations, parang, wvs, astr_hdrs, filt_band, fpm_band, ppm_band, spot_fluxes, prihdr, exthdr


def generate_psf(frame, locations, boxrad=5, medianboxsize=30):
    """
    Generates a P1640 PSF for the frame based on the satellite spots

    Args:
        frame: 2d frame of data
        location: array of (N,2) containing [x,y] coordinates of all N satellite spots
        boxrad: half length of box to use to pull out PSF
        medianboxsize: size in pixels of box for median filter

    Returns:
        genpsf: 2d frame of size (2*boxrad+1, 2*boxrad+1) with average PSF of satellite spots
    """
    genpsf = []
    #mask nans
    cleaned = np.copy(frame)
    cleaned[np.where(np.isnan(cleaned))] = 0
    #mask source for median filter
    masked = np.copy(cleaned)
    for loc in locations:
        spotx = np.int(np.round(loc[0]))
        spoty = np.int(np.round(loc[1]))
        masked[spotx-boxrad:spotx+boxrad+1, spoty-boxrad:spoty+boxrad+1] = np.nanmedian(
            masked.reshape(masked.shape[0]*masked.shape[1]))
    #subtract out median filtered image

    #cleaned -= ndimage.median_filter(masked, size=(medianboxsize,medianboxsize))

    for loc in locations:
        #grab satellite spot positions
        spotx = loc[0]
        spoty = loc[1]

        #interpolate image to grab satellite psf with it centered
        #add .1 to make sure we get 2*boxrad+1 but can't add +1 due to floating point precision (might cause us to
        #create arrays of size 2*boxrad+2)
        x,y = np.meshgrid(np.arange(spotx-boxrad, spotx+boxrad+0.1, 1), np.arange(spoty-boxrad, spoty+boxrad+0.1, 1))
        spotpsf = ndimage.map_coordinates(cleaned, [y,x])
        genpsf.append(spotpsf)

    genpsf = np.array(genpsf)
    genpsf = np.mean(genpsf, axis=0) #average the different psfs together
    return genpsf


def rescale_wvs(exthdrs, wvs, refwv=18, skipslices=None):
    """
    Hack to try to fix wavelength scaling issue. This will calculate the scaling between channels,
    and adjust the wavelength solution such that the scaling comes out linear in scaling vs wavelength.
    Finicky - requires that all images in the dataset have the same number of wavelength channels
    Args:
        exthdrs: a list of extension headers, from a pyklip.instrument dataset
        wvs: a list of wvs (can repeat. This function will only look at the first cube's wavelenghts)
        refwv (optional): integer index of the channel to normalize the scaling
        skipslices: list of skipped wavelength slices (needs to be consistent with the ones skipped by wv)
    Returns:
        scaled_wvs: Nlambda*Nexthdrs array of wavelengths that produce a linear plot of wavelength vs scaling
    """
    #wvs_mean = wvs.reshape(len(exthdrs), len(wvs)/len(exthdrs)).mean(axis=0)
    wv_indicies = range(0, exthdrs[0]['NAXIS3'])
    if skipslices is not None:
        wv_indicies = np.delete(wv_indicies, skipslices)
    sats = np.array([[[h['SATS{0}_{1}'.format(i,j)].split() for i in wv_indicies]
                          for j in range(0,4)] for h in exthdrs], dtype=np.float)
    sats = sats.mean(axis=0)
    pairs = [(0,3), (1,2)]
    separations = np.mean([0.5*np.sqrt(np.diff(sats[p,:,0], axis=0)[0]**2 + np.diff(sats[p,:,1], axis=0)[0]**2) 
                           for p in pairs], 
                          axis=0) # average over each pair, the first axis
    scaling_factors = separations/separations[refwv]
    scaled_wvs = scaling_factors*wvs[refwv]
    return np.tile(scaled_wvs, len(exthdrs))





__author__ = 'JB'

from scipy.signal import correlate2d
import numpy as np
from copy import copy

def calculate_cc(image, PSF,spectrum = None, nans2zero=False):
    """
    Perform a cross correlation on the current loaded file.

    Args:
        image: image to get the cross correlation from.
        PSF: Template for the cross correlation
        spectrum: If not None, combine the 3D cube into a 2D image using a weighted mean with spectrum.
        nans2zero: If True, temporarily replace all nans values with zeros for the cross correlation

    Return: cross correlated image.
    """
    if np.size(image.shape) == 2:
        ny,nx = image.shape
    if np.size(image.shape) == 3:
        nl,ny,nx = image.shape

    if np.size(PSF.shape) == 2:
        ny_PSF,nx_PSF = PSF.shape
    if np.size(PSF.shape) == 3:
        nl_PSF,ny_PSF,nx_PSF = PSF.shape

    image_cpy = copy(image)

    if spectrum is not None:
        image_collapsed = np.zeros((ny,nx))
        for k in range(nl):
            image_collapsed = image_collapsed + spectrum[k]*image_cpy[k,:,:]
        image_cpy = image_collapsed/np.sum(spectrum)

    if nans2zero:
        where_nans = np.where(np.isnan(image_cpy))
        image_cpy = np.nan_to_num(image_cpy)

    # We have to make sure the PSF dimensions are odd because correlate2d shifts the image otherwise...
    if (nx_PSF % 2 ==0):
        PSF_tmp = np.zeros((ny_PSF,nx_PSF+1))
        PSF_tmp[0:ny_PSF,0:nx_PSF] = PSF
        PSF = PSF_tmp
        nx_PSF = nx_PSF +1
    if (ny_PSF % 2 ==0):
        PSF_tmp = np.zeros((ny_PSF+1,nx_PSF))
        PSF_tmp[0:ny_PSF,0:nx_PSF] = PSF
        PSF = PSF_tmp
        ny_PSF = ny_PSF +1

    # Check if the input file is 2D or 3D
    if np.size(image_cpy.shape) == 3: # If the file is a 3D cube
        image_convo = np.zeros(image_cpy.shape)
        for l_id in np.arange(nl):
            image_convo[l_id,:,:] = correlate2d(image_cpy[l_id,:,:],PSF,mode="same")
    else: # image is 2D
        image_convo = correlate2d(image_cpy,PSF,mode="same")

    if nans2zero:
        image_convo[where_nans] = np.nan

    return image_convo



import os
import glob
from copy import copy

import numpy as np
from scipy import optimize
import scipy.ndimage as ndimage
import scipy.interpolate as interp
from scipy.optimize import minimize
from scipy.interpolate import interp1d
import astropy.modeling as modeling



import pyklip.spectra_management as spec


def convert_pa_to_image_polar(pa, astr_hdr):
    """
    Given a position angle (angle to North through East), calculate what
    polar angle theta (angle from +X CCW towards +Y) it corresponds to

    Args:
        pa: position angle in degrees
        astr_hdr: wcs astrometry header (astropy.wcs)

    Returns:
        theta: polar angle in degrees
    """
    rot_det = astr_hdr.wcs.cd[0,0] * astr_hdr.wcs.cd[1,1] - astr_hdr.wcs.cd[0,1] * astr_hdr.wcs.cd[1,0]
    if rot_det < 0:
        rot_sgn = -1.
    else:
        rot_sgn = 1.
    # calculate CCW rotation from +Y to North in radians
    rot_YN = np.arctan2(rot_sgn * astr_hdr.wcs.cd[0,1],rot_sgn * astr_hdr.wcs.cd[0,0])
    # now that we know where north it,
    # find the CCW rotation from +Y to find location of planet
    rot_YPA = rot_YN - rot_sgn*pa*np.pi/180. #radians
    # rot_YPA = rot_YN + pa*np.pi/180. #radians

    theta = rot_YPA * 180./np.pi + 90.0 #degrees
    return theta


def convert_polar_to_image_pa(theta, astr_hdr):
    """
    Reversed engineer from covert_pa_to_image_polar by JB. Actually JB doesn't quite understand how it works...

    Args:
        theta: parallactic angle in degrees
        astr_hdr: wcs astrometry header (astropy.wcs)

    Returns:
        theta: polar angle in degrees
    """
    rot_det = astr_hdr.wcs.cd[0,0] * astr_hdr.wcs.cd[1,1] - astr_hdr.wcs.cd[0,1] * astr_hdr.wcs.cd[1,0]
    if rot_det < 0:
        rot_sgn = -1.
    else:
        rot_sgn = 1.
    #calculate CCW rotation from +Y to North in radians
    rot_YN = np.arctan2(rot_sgn * astr_hdr.wcs.cd[0,1],rot_sgn * astr_hdr.wcs.cd[0,0])

    rot_YPA = (theta-90)*np.pi/180.

    pa = rot_sgn*(rot_YN-rot_YPA)* 180./np.pi

    return pa


def _inject_gaussian_planet(frame, xpos, ypos, amplitude, fwhm=3.5, stampsize=None, field_dependent_correction=None, center=None):
    """
    Injects a fake planet with a Gaussian PSF into a dataframe

    Args:
        frame: a 2D data frame
        xpos,ypos: x,y location (in pixels) where the planet should be
        amplitude: peak of the Gaussian PSf (in appropriate units not dictacted here)
        fwhm: fwhm of gaussian
        stampsize: integer specfying the width of the stamp box in which the planet is defined
        field_dependent_correction: a function of the form ``output_stamp = correction(input_stamp, input_dx, input_dy)
                            where, input_stamp is a 2-D image of shape (y_stamp, x_stamp). input_dx and input_dy have
                            the same shape and represent the offset of each pixel from the star (in pixel units). The 
                            function returns an output_stamp of the same shape, but corrected for any field dependent
                            throughputs or distortions.
        center: (x,y) reference position (typically of a coronagraph center) that is used ONLY by the field dependent correction
    Returns:
        frame: the frame with the injected planet
    """
    if stampsize is None:
        stampsize = 3 * fwhm
    # convert boxsize to an integer
    stampsize = int(np.ceil(stampsize))

    # figure out sigma when given FWHM
    sigma = fwhm/(2.*np.sqrt(2*np.log(2)))

    # create a coordinate system for the PSF centered about the closest pixel to the planet
    y, x = np.indices([stampsize, stampsize])
    y -= stampsize // 2 # center about 0
    x -= stampsize // 2  # center about 0
    # find nearest pixel to center coordinate around
    x_int = int(round(xpos))
    y_int = int(round(ypos))
    x += x_int
    y += y_int

    xmin = x[0][0]
    xmax = x[-1][-1]
    ymin = y[0][0]
    ymax = y[-1][-1]

    psf = amplitude * np.exp(-((x - xpos)**2. + (y - ypos)**2.) / (2. * sigma**2))
    
    # if specified, make field dependent PSF correction
    if field_dependent_correction is not None:
        dx_psf = np.arange(xmin, xmax+1, 1) - center[0]
        dy_psf = np.arange(ymin, ymax+1, 1) - center[1]
        # convert to 2D
        dx_psf, dy_psf = np.meshgrid(dx_psf, dy_psf)
        psf = field_dependent_correction(psf, dx_psf, dy_psf)

    frame[ymin:ymax+1, xmin:xmax+1] += psf
    return frame


def inject_planet(frames, centers, inputflux, astr_hdrs, radius, pa, fwhm=3.5, thetas=None, stampsize=None, field_dependent_correction=None):
    """
    Injects a fake planet into a dataset either using a Gaussian PSF or an input PSF

    Args:
        frames: array of (N,y,x) for N is the total number of frames
        centers: array of size (N,2) of [x,y] coordiantes of the image center
        inputflux: EITHER array of size N of the peak flux of the fake planet in each frame (will inject a Gaussian PSF)
                   OR array of size (N,psfy,psfx) of template PSFs. The brightnesses should be scaled and the PSFs
                   should be centered at the center of each of the template images
        astr_hdrs: array of size N of the WCS headers
        radius: separation of the planet from the star
        pa: position angle (in degrees) of  planet
        fwhm: fwhm (in pixels) of gaussian
        thetas: ignore PA, supply own thetas (CCW angle from +x axis toward +y)
                array of size N
        stampsize: in pixels, the width of the square stamp to inject the image into. Defaults to 3*fwhm if None
        field_dependent_correction: a function of the form ``output_stamp = correction(input_stamp, input_dx, input_dy)
                                    where, input_stamp is a 2-D image of shape (y_stamp, x_stamp). input_dx and input_dy have
                                    the same shape and represent the offset of each pixel from the star (in pixel units). The 
                                    function returns an output_stamp of the same shape, but corrected for any field dependent
                                    throughputs or distortions.

    Returns:
        saves result in input "frames" variable
    """

    if thetas is None:
        thetas = np.array([convert_pa_to_image_polar(pa, astr_hdr) for astr_hdr in astr_hdrs])

    if (np.size(inputflux) == 1):
        #input is probably a number and we want an array
        inputflux = np.ones(frames.shape[0]) * inputflux

    for frame, center, inputpsf, theta in zip(frames, centers, inputflux, thetas):
        #calculate the x,y location of the planet for each image
        #theta = covert_pa_to_image_polar(pa, astr_hdr)
        x_pl = radius * np.cos(theta*np.pi/180.) + center[0]
        y_pl = radius * np.sin(theta*np.pi/180.) + center[1]

        ny,nx = frame.shape

        #now that we found the planet location, inject it
        #check whether we are injecting a gaussian of a template PSF
        if isinstance(inputpsf, np.ndarray):
            # stampsize should defautl to minimum dimension of PSF
            if stampsize is None:
                stampsize = int(np.min(inputpsf.shape))
            # convert boxsize to an integer
            stampsize = int(np.ceil(stampsize))

            #shift psf so that center is aligned
            #calculate center of box
            boxsize = inputpsf.shape[0]
            #Using JB's array convention instead.
            boxcent = boxsize // 2

            # create a coordinate system for the PSF centered about the closest pixel to the planet
            ystamp, xstamp = np.indices([stampsize, stampsize])
            ystamp -= stampsize // 2  # center about 0
            xstamp -= stampsize // 2  # center about 0
            # find nearest pixel to center coordinate around
            x_int = int(round(x_pl))
            y_int = int(round(y_pl))
            xstamp += x_int
            ystamp += y_int
            # index bounds
            xmin = xstamp[0][0]
            xmax = xstamp[-1][-1]
            ymin = ystamp[0][0]
            ymax = ystamp[-1][-1]

            if xmin>= nx or ymin>= ny or xmax<= 0 or ymax<= 0:
                continue
            # find corresponding pixels in the PSF
            xpsf = xstamp - x_pl + boxcent
            ypsf = ystamp - y_pl + boxcent

            # Crop the edge if injection at the edge of the image
            if xmin < 0:
                ypsf = ypsf[:,-xmin::]
                xpsf = xpsf[:,-xmin::]
                xmin=0
            if ymin < 0:
                ypsf = ypsf[-ymin::,:]
                xpsf = xpsf[-ymin::,:]
                ymin = 0
            if xmax >= nx:
                ypsf = ypsf[:,:-(xmax-nx + 1)]
                xpsf = xpsf[:,:-(xmax-nx + 1)]
                xmax = nx - 1
            if ymax >= ny:
                ypsf = ypsf[:-(ymax-ny + 1),:]
                xpsf = xpsf[:-(ymax-ny + 1),:]
                ymax = ny - 1

            psf = ndimage.map_coordinates(inputpsf, [ypsf, xpsf], mode='constant', cval=0.0)
            # if specified, make field dependent PSF correction
            if field_dependent_correction is not None:
                dx_psf = np.arange(xmin, xmax+1, 1) - center[0]
                dy_psf = np.arange(ymin, ymax+1, 1) - center[1]
                # convert to 2D
                dx_psf, dy_psf = np.meshgrid(dx_psf, dy_psf)
                psf = field_dependent_correction(psf, dx_psf, dy_psf)

            #inject into frame
            frame[ymin:ymax + 1, xmin:xmax + 1] += psf
        else:
            if stampsize is None:
                stampsize = 3 * fwhm
            # convert boxsize to an integer
            stampsize = int(np.ceil(stampsize))

            _inject_gaussian_planet(frame, x_pl, y_pl, inputpsf, fwhm=fwhm, stampsize=stampsize, field_dependent_correction=field_dependent_correction, center=center)


def generate_dataset_with_fakes(dataset, fake_position_dict, fake_flux_dict, spectrum = None, PSF_cube = None, PSF_cube_wvs = None,
                                star_type = None, mute = False, SpT_file_csv = None, real_planets_pos = None, sep_skip_real_pl = None,
                                pa_skip_real_pl = None,dn_per_contrast=None,star_spec = None):
    '''
    Generate spectral datacubes with fake planets.
    It will do a copy of the cubes read in GPIData after having injected fake planets in them.
    This new set of cubes can then be reduced in the same manner as the campaign data.

    Doesn't work with remove slice: assumes that the dataset is made of a list of similar datacubes or images.

    Args:
        dataset: An object of type GPIData.
                The fakes are injected directly into dataset so you should make a copy of dataset prior to running this
                function.
                In order for the function to query simbad for the spectral type of the star, the attribute object_name needs
                to be defined in dataset.
        fake_position_dict:
                Dictionary defining the way the fake planets are positionned
                - fake_position_dict["mode"]="sector": Put a planet in each klip sector. Can actually generate several
                        datasets in which the planets will be shifted in separation and position angle with respect to one
                        another.
                        It can be usefull for fake based contrast curve calculation.
                        Several parameters needs to be defined.
                        - fake_position_dict["annuli"]: Number of annulis in the image
                        - fake_position_dict["subsections"]: Number of angular sections in the image
                        - fake_position_dict["sep_shift"]: separation shift from the center of the sectors
                        - fake_position_dict["pa_shift"]: position angle shift from the center of the sectors
                - fake_position_dict["mode"]="custom": Put planets at given (separation, position angle).
                        The following parameter needs to be defined
                        - fake_position_dict["pa_sep_list"]: List of tuple [(r1,pa1),(r2,pa2),...] with each tuple giving
                                the separation and position angle of each planet to be injected.
                - fake_position_dict["mode"]="ROC": Generate fake for ROC curves calculation. Use hard-coded parameters.
            fake_flux_dict:
                Dictionary defining the way in which the flux of the fake is defined.
                - fake_flux_dict["mode"]="contrast": Defines the contrast value of the fakes.
                        - fake_flux_dict["contrast"]: Contrast of the fake planets
                - fake_flux_dict["mode"]="SNR": Defines the brightness of the fakes relatively to the satellite spots.
                        - fake_flux_dict["SNR"]: SNR wished for the fake planets.
                        - fake_flux_dict["sep_arr"]: Separation sampling of the contrast curve in pixels.
                        - fake_flux_dict["contrast_arr"]: 5 sigma contrast curve (planet to star ratio).
        PSF_cube: the psf of the image. A numpy array with shape (wv, y, x)
        PSF_cube_wvs: the wavelegnths that correspond to the input psfs
        spectrum: spectrum name (string) or array
                    - "host_star_spec": The spectrum from the star or the satellite spots is directly used.
                                        It is derived from the inverse of the calibrate_output() output.
                    - "constant": Use a constant spectrum np.ones(nl).
                    - other strings: name of the spectrum file in #pykliproot#/spectra/*/ with pykliproot the
                                    directory in which pyklip is installed. It that case it should be a spectrum
                                    from Mark Marley or one following the same convention.
                                    Spectrum will be corrected for transmission.
                    - ndarray: 1D array with a user defined spectrum. Spectrum will be corrected for transmission.
        star_type: Spectral type of the current star. If None, Simbad is queried.
        mute: If True prevent printed log outputs.
        suffix: Suffix to be added at the end of the filenames.
        SpT_file_csv: Filename of the table (.csv) contaning the spectral type of the stars.
        real_planets_pos: list of position of real point sources in the dataset that should be avoided when injecting fakes.
                        [(sep1,pa1),(sep2,pa2),...] with the separation in pixels and the position angle in degrees.
        sep_skip_real_pl: Limit in seperation of how close a fake can be injected of a known GOI.
        pa_skip_real_pl: Limit in position angle  of how close a fake can be injected of a known GOI.
        dn_per_contrast: array of the same size as spectrum giving the conversion between the peak flux of a planet in
                        data number and its contrast.
        star_spec: 1D array stellar spectrum sampling dataset.wvs. Otherwise uses a pickles spectrum in which the
                   temperature in interpolated from the spectral type.

    '''

    if sep_skip_real_pl is None:
        sep_skip_real_pl = 20
    if pa_skip_real_pl is None:
        pa_skip_real_pl = 90

    try:
        star_name = dataset.object_name.replace(" ","_")
    except:
        star_name = "noname"

    nl, ny, nx = dataset.input.shape
    if dn_per_contrast is None:
        dn_per_contrast = 1./dataset.calibrate_output(np.ones((nl,1,1)),spectral=True).squeeze()


    # Make sure the total flux of each PSF is unity for all wavelengths
    # So the peak value won't be unity.
    # print("np.sum(PSF_cube)",np.sum(PSF_cube))
    PSF_cube = PSF_cube/np.nansum(PSF_cube,axis=(1,2))[:,None,None]
    # print("np.sum(PSF_cube)",np.sum(PSF_cube))
    # Get the conversion factor from peak spectrum to aperture based spectrum
    aper_over_peak_ratio = 1/np.nanmax(PSF_cube,axis=(1,2))
    aper_over_peak_ratio_tiled = np.zeros(nl)#wavelengths
    for k,wv in enumerate(dataset.wvs):
        aper_over_peak_ratio_tiled[k] = aper_over_peak_ratio[spec.find_nearest(PSF_cube_wvs,wv)[1]]
    # Summed DN flux of the star in the entire dataset calculated from dn_per_contrast
    host_star_spec = aper_over_peak_ratio_tiled*dn_per_contrast
    star_flux = np.sum(host_star_spec)
    # print(star_flux,aper_over_peak_ratio_tiled[0],dn_per_contrast[0],aper_over_peak_ratio)
    # # exit()
    host_star_spec = host_star_spec/np.mean(host_star_spec)
    nl_psf, ny_psf, nx_psf = PSF_cube.shape
    inputpsfs = np.zeros((nl,ny_psf,nx_psf))
    for k,wv in enumerate(dataset.wvs):
        inputpsfs[k,:,:] = PSF_cube[spec.find_nearest(PSF_cube_wvs,wv)[1],:,:]


    if np.size(np.unique(dataset.wvs)) == 1:
        star_sp = np.ones(dn_per_contrast.shape)
    else:
        if star_spec is None:
            if star_type is None:
                star_type = spec.get_specType(star_name,SpT_file_csv)
            # Interpolate a spectrum of the star based on its spectral type/temperature
            wv,star_sp = spec.get_star_spectrum(dataset.wvs,star_type)
        else:
            star_sp = star_spec

    # Define the output Foldername
    if isinstance(spectrum, str):

        # Do the best it can with the spectral information given in inputs.
        if spectrum == "host_star_spec":
            # If spectrum_filename is an empty string the function takes the sat spot spectrum by default.
            if not mute:
                print("Default host star specrum will be used.")
            spectrum_vec = copy(host_star_spec)
            spectrum_name = "host_star_spec"
        elif spectrum == "constant":
            if not mute:
                print("Spectrum is not or badly defined so taking flat spectrum")
            spectrum_vec = np.ones(nl)
            spectrum_name = "constant"
        else :
            pykliproot = os.path.dirname(os.path.realpath(spec.__file__))
            spectrum_filename = os.path.abspath(glob.glob(os.path.join(pykliproot,"spectra","*",spectrum+".flx"))[0])
            spectrum_name = spectrum_filename.split(os.path.sep)
            spectrum_name = spectrum_name[len(spectrum_name)-1].split(".")[0]

            # spectrum_filename is not empty it is assumed to be a valid path.
            if not mute:
                print("Spectrum model: "+spectrum_filename)
            # Interpolate the spectrum of the planet based on the given filename
            wv,planet_sp = spec.get_planet_spectrum(spectrum_filename,dataset.wvs)

            # Correct the ideal spectrum given in spectrum_filename for atmospheric and instrumental absorption.
            spectrum_vec = (host_star_spec/star_sp)*planet_sp

    elif isinstance(spectrum, np.ndarray):
        planet_sp = spectrum
        spectrum_name = "custom"

        # Correct the ideal spectrum given in spectrum_filename for atmospheric and instrumental absorption.
        spectrum_vec = (host_star_spec/star_sp)*planet_sp
    else:
        raise ValueError("Invalid spectrum: {0}".format(spectrum))

    spectrum_vec = spectrum_vec/np.mean(spectrum_vec)

    if real_planets_pos is not None:
        sep_real_object_list = [sep for (sep,pa) in real_planets_pos] # in pixels
        pa_real_object_list = [pa for (sep,pa) in real_planets_pos] # in degrees

    # inputflux_is_def = False
    # Build the list of separation and position angles for the fakes
    if fake_position_dict["mode"] == "custom":
        sep_pa_iter_list = fake_position_dict["pa_sep_list"]

    if fake_position_dict["mode"] == "spirals":
        if "pa_shift" in fake_position_dict:
            pa_shift = fake_position_dict["pa_shift"]
        else:
            pa_shift = 0.0
        # Calculate the radii of the annuli like in klip_adi_plus_sdi using the first image
        # We want to inject one planet per section where klip is independently applied.
        if "annuli" in fake_position_dict:
            annuli = fake_position_dict["annuli"]
        else:
            annuli = 8
        if "dr" in fake_position_dict:
            dr = fake_position_dict["dr"]
        else:
            dr = 15.
        delta_th = 90

        # Get parallactic angle of where to put fake planets
        # PSF_dist = 20 # Distance between PSFs. Actually length of an arc between 2 consecutive PSFs.
        # delta_pa = 180/np.pi*PSF_dist/radius
        pa_list = np.arange(-180.,180.-0.01,delta_th) + pa_shift
        radii_list = np.array([dr * annuli_it + dataset.IWA + 2.5 for annuli_it in range(annuli)])
        pa_grid, radii_grid = np.meshgrid(pa_list,radii_list)
        # for row_id in range(pa_grid.shape[0]):
        #     pa_grid[row_id,:] = pa_grid[row_id,:] + 30
        for col_id in range(radii_grid.shape[1]):
            radii_grid[:,col_id] = radii_grid[:,col_id] + dr/4*np.mod(col_id,4)
        pa_grid[range(1,annuli,3),:] += 30
        pa_grid[range(2,annuli,3),:] += 60
        pa_grid = pa_grid + 5

        sep_pa_iter_list = [(r, p) for r_arr, p_arr in zip(radii_grid, pa_grid) for r, p in zip(r_arr, p_arr)]

    if fake_position_dict["mode"] == "sector":
        annuli = fake_position_dict["annuli"]
        subsections = fake_position_dict["subsections"]
        sep_shift = fake_position_dict["sep_shift"]
        pa_shift = fake_position_dict["pa_shift"]

        # Calculate the radii of the annuli like in klip_adi_plus_sdi using the first image
        # We want to inject one planet per section where klip is independently applied.
        dims = dataset.input.shape
        x_grid, y_grid = np.meshgrid(np.arange(dims[2] * 1.0), np.arange(dims[1] * 1.0))
        nanpix = np.where(np.isnan(dataset.input[0]))
        if np.size(nanpix) == 0:
            OWA = np.sqrt(np.max((x_grid) ** 2 + (y_grid) ** 2))
        else:
            OWA = np.sqrt(np.min((x_grid[nanpix] - dataset.centers[0][0]) ** 2 + (y_grid[nanpix] - dataset.centers[0][1]) ** 2))
        dr = float(OWA - dataset.IWA) / (annuli)
        delta_th = 360./subsections

        # Get parallactic angle of where to put fake planets
        # PSF_dist = 20 # Distance between PSFs. Actually length of an arc between 2 consecutive PSFs.
        # delta_pa = 180/np.pi*PSF_dist/radius
        pa_list = np.arange(-180.,180.-0.01,delta_th) + pa_shift
        radii_list = np.array([dr * annuli_it + dataset.IWA + dr/2.for annuli_it in range(annuli-1)]) + sep_shift
        pa_grid, radii_grid = np.meshgrid(pa_list,radii_list)
        pa_grid[range(1,annuli-1,2),:] += delta_th/2.

        sep_pa_iter_list = [(r, p) for r_arr, p_arr in zip(radii_grid, pa_grid) for r, p in zip(r_arr, p_arr)]

    # fake_flux_dict = dict(mode = "SNR",sep_arr = sep_samples, contrast_arr=Ttype_contrast)
    if (fake_flux_dict["mode"] == "contrast"):
        if isinstance(fake_flux_dict["contrast"], list) or isinstance(fake_flux_dict["contrast"], np.ndarray):
            planets_contrasts = fake_flux_dict["contrast"]
        else:
            planets_contrasts = [fake_flux_dict["contrast"],]*len(sep_pa_iter_list)
    elif (fake_flux_dict["mode"] == "SNR"):
        sep_arr = np.array(fake_flux_dict["sep_arr"])
        cont_arr = np.array(fake_flux_dict["contrast_arr"])
        f = interp1d(sep_arr[np.where(np.isfinite(cont_arr))], cont_arr[np.where(np.isfinite(cont_arr))],
                     bounds_error=False,fill_value=np.nanmin(fake_flux_dict["contrast_arr"]))
        planets_contrasts = [fake_flux_dict["SNR"]*f(sep)/5. for (sep,pa) in sep_pa_iter_list]

    extra_keywords = {}
    # Loop for injecting fake planets. One planet per section of the image.
    for fake_id, ((radius,pa),contrast) in enumerate(zip(sep_pa_iter_list,planets_contrasts)):
        x_max_pos = radius*np.cos(np.radians(90+pa))
        y_max_pos = radius*np.sin(np.radians(90+pa))

        # Not injecting the planet if too close to real object
        if real_planets_pos is not None:
            too_close = False
            for sep_real_object,pa_real_object  in zip(sep_real_object_list,pa_real_object_list):
                delta_angle = np.min([np.abs(np.mod(pa,360)-np.mod(pa_real_object,360)),
                               np.min([np.mod(pa,360),np.mod(pa_real_object,360)])+360-np.max([np.mod(pa,360),np.mod(pa_real_object,360)])])
                if np.abs(sep_real_object-radius) < sep_skip_real_pl and delta_angle < pa_skip_real_pl:
                    too_close = True
                    if not mute:
                        print("Skipping planet. Real object too close.")
                    break
            if too_close:
                continue


        spectrum_corr = spectrum_vec/np.sum(spectrum_vec)*star_flux*contrast
        inputpsfs = inputpsfs/np.nansum(inputpsfs,axis=(1,2))[:,None,None]
        inputpsfs = inputpsfs*spectrum_corr[:,None,None]

        try:
            if not mute:
                print("injecting planet position ("+str(radius)+"pix,"+str(pa)+"degree)")
            # inject fake planet at given radius,pa into dataset.input
            inject_planet(dataset.input, dataset.centers, inputpsfs, dataset.wcs, radius, pa,
                          stampsize=np.min([ny_psf, nx_psf]))#, thetas=pa+dataset.PAs)

            # Save fake planet position in headers
            extra_keywords["FKPA{0:02d}".format(fake_id)] = pa
            extra_keywords["FKSEP{0:02d}".format(fake_id)] = radius
            extra_keywords["FKCONT{0:02d}".format(fake_id)] = contrast
            extra_keywords["FKPOSX{0:02d}".format(fake_id)] = x_max_pos
            extra_keywords["FKPOSY{0:02d}".format(fake_id)] = y_max_pos
            extra_keywords["FKSPEC{0:02d}".format(fake_id).format(fake_id)] = spectrum_name
        except OverflowError:
            pass
        # except:
        #     if not mute:
        #         print("Failed to inject planet position ("+str(radius)+"pix,"+str(pa)+"degree)")

    return dataset,extra_keywords


def _construct_gaussian_disk(x0,y0, xsize,ysize, intensity, angle, fwhm=3.5):
    """
    Constructs a rectangular slab for a disk with a vertical gaussian profile

    Args:
        x0,y0: center of disk
        xsize, ysize: x and y dimensions of the output image
        intensity: peak intensity of the disk (whatever units you want)
        angle: orientation of the disk plane (CCW from +x axis) [degrees]
        fwhm: FWHM of guassian profile (in pixels)

    Returns:
        disk_img: 2d array of size (ysize,xsize) with the image of the disk
    """

    #construct a coordinate system
    x,y = np.meshgrid(np.arange(xsize*1.0), np.arange(ysize*1.0))

    #center at image center
    x -= x0
    y -= y0

    #rotate so x is parallel to the disk plane, y is vertical cuts through the disk
    #so need to do a CW rotation
    rad_angle = angle * np.pi/180.
    xp = x * np.cos(rad_angle) + y * np.sin(rad_angle) + x0
    yp = -x * np.sin(rad_angle) + y * np.cos(rad_angle) + y0

    sigma = fwhm/(2 * np.sqrt(2*np.log(2)))
    disk_img = intensity / (np.sqrt(2*np.pi) * sigma) * np.exp(-(yp-y0)**2/(2*sigma**2))

    return disk_img


def inject_disk(frames, centers, inputfluxes, astr_hdrs, pa, fwhm=3.5):
    """
    Injects a fake disk into a dataset

    Args:
        frames: array of (N,y,x) for N is the total number of frames
        centers: array of size (N,2) of [x,y] coordiantes of the image center
        intputfluxes: array of size N of the peak flux of the fake disk in each frame OR
                      array of 2-D models (North up East left) to inject into the data.
                            (Disk is assumed to be centered at center of image)
        astr_hdrs: array of size N of the WCS headers
        pa: position angles angle (in degrees) of disk plane
        fwhm: if injecting a Gaussian disk (i.e inputfluxes is an array of floats), fwhm of Gaussian

    Returns:
        saves result in input "frames" variable
    """

    for frame, center, inputpsf, astr_hdr in zip(frames, centers, inputfluxes, astr_hdrs):
        #calculate the rotation angle in the pixel plane
        theta = convert_pa_to_image_polar(pa, astr_hdr)

        if isinstance(inputpsf, np.ndarray):
            # inject real data
            # rotate and grab pixels of disk that can be injected into the image
            # assume disk is centered
            xpsf0 = inputpsf.shape[1]/2
            ypsf0 = inputpsf.shape[0]/2
            #grab the pixel numbers for the data
            ximg, yimg = np.meshgrid(np.arange(frame.shape[1]), np.arange(frame.shape[0]))
            #rotate them to extract the disk at the right angle
            ximg -= center[0]
            yimg -= center[1]
            theta_rad = np.radians(theta)
            ximgp = ximg * np.cos(theta_rad) + yimg * np.sin(theta_rad) + xpsf0
            yimgp = -ximg * np.sin(theta_rad) + yimg * np.cos(theta_rad) + ypsf0
            #interpolate and inject datqa
            frame += ndimage.map_coordinates(inputpsf, [yimgp, ximgp])
        else:
            #inject guassian bar into data
            frame += _construct_gaussian_disk(center[0], center[1], frame.shape[1], frame.shape[0], inputpsf, theta, fwhm=fwhm)


def gauss2d(x0, y0, peak, sigma):
    """
    2d symmetric guassian function for guassfit2d

    Args:
        x0,y0: center of gaussian
        peak: peak amplitude of guassian
        sigma: stddev in both x and y directions
    """
    sigma *= 1.0
    return lambda y,x: peak*np.exp( -(((x-x0)/sigma)**2+((y-y0)/sigma)**2)/2)


def gaussfit2d(frame, xguess, yguess, searchrad=5, guessfwhm=3, guesspeak=1, refinefit=True):
    """
    Fits a 2d gaussian to the data at point (xguess, yguess)

    Args:
        frame: the data - Array of size (y,x)
        xguess,yguess: location to fit the 2d guassian to (should be pretty accurate)
        searchrad: 1/2 the length of the box used for the fit
        guessfwhm: approximate fwhm to fit to
        guesspeak: approximate flux
        refinefit: whether to refine the fit of the position of the guess

    Returns:
        peakflux: the peakflux of the gaussian
        fwhm: fwhm of the PFS in pixels
        xfit: x position (only chagned if refinefit is True)
        yfit: y position (only chagned if refinefit is True)
    """
    if not isinstance(searchrad, int):
        raise ValueError("searchrad needs to be an integer")

    x0 = np.rint(xguess).astype(int)
    y0 = np.rint(yguess).astype(int)
    #construct our searchbox
    fitbox = np.copy(frame[y0-searchrad:y0+searchrad+1, x0-searchrad:x0+searchrad+1])

    #mask bad pixels
    fitbox[np.where(np.isnan(fitbox))] = 0
 
    #fit a least squares gaussian to refine the fit on the source, otherwise just use the guess
    if refinefit:
        #construct the residual to the fit
        errorfunction = lambda p: np.ravel(gauss2d(*p)(*np.indices(fitbox.shape)) - fitbox)
   
        #do a least squares fit. Note that we use searchrad for x and y centers since we're narrowed it to a box of size
        #(2searchrad+1,2searchrad+1)

        guess = (searchrad, searchrad, guesspeak, guessfwhm/(2 * np.sqrt(2*np.log(2))))

        p, success = optimize.leastsq(errorfunction, guess)

        xfit = p[0]
        yfit = p[1]
        peakflux = p[2]
        fwhm = p[3] * (2 * np.sqrt(2*np.log(2)))
    else:
        xfit = xguess-x0 + searchrad
        yfit = yguess-y0 + searchrad
        fwhm = guessfwhm
   
    #ok now, to really calculate fwhm and flux, because we really need that right, we're going
    # to use what's in the GPI DRP pipeline to measure satellite spot fluxes instead of
    # a least squares gaussian fit. Apparently my least squares fit relatively underestimates
    # the flux so it's not consistent.
    # grab a radial profile of the fit
    rs = np.linspace(0, searchrad+1, np.max([15, searchrad+1])) # should always have at least pixel resolution
    thetas = np.arange(0,2*np.pi, 1./searchrad) #divide maximum circumfrence into equal parts
    radprof = [np.mean(ndimage.map_coordinates(fitbox, [thisr*np.sin(thetas)+yfit, thisr*np.cos(thetas)+xfit])) for thisr in rs]
    #now interpolate this radial profile to get fwhm
    try:
        radprof_interp = interp.interp1d(radprof, rs)
        fwhm = 2*radprof_interp(np.max(radprof[0])/2)
    except ValueError:
        # use old FWHM
        pass


    #now calculate flux
    xfitbox, yfitbox = np.meshgrid(np.arange(0,fitbox.shape[1], 1.0)-xfit, np.arange(0, fitbox.shape[0], 1.0)-yfit)
    #correlate data with a gaussian to get flux
    sigma = fwhm/(2*np.sqrt(2*np.log(2)))
    ## attempt to calculate sigma using moments
    #sigmax = np.sqrt(np.nansum(xfitbox*xfitbox*fitbox)/np.nansum(fitbox) - (np.nansum(xfitbox*fitbox)/np.nansum(fitbox))**2)
    #sigmay = np.sqrt(np.nansum(yfitbox*yfitbox*fitbox)/np.nansum(fitbox) - (np.nansum(yfitbox*fitbox)/np.nansum(fitbox))**2)
    #sigma = np.nanmean([sigmax, sigmay])
    #print(sigma, sigmax, sigmay)
    gmask = np.exp(-(xfitbox**2+yfitbox**2)/(2.*sigma**2))
    outofaper = np.where(xfitbox**2 + yfitbox**2 > searchrad**2)
    gmask[outofaper] = 0 
    corrflux = np.nansum(fitbox*gmask)/np.sum(gmask*gmask)

    if not np.isfinite(corrflux):
        # if it's infinite, it is bad
        corrflux = np.nan

    # convert xfit, yfit back to image coordinates
    xfit = xfit - searchrad + x0
    yfit = yfit - searchrad + y0

    return corrflux, fwhm, xfit, yfit

def airyfit2d(frame, xguess, yguess, searchrad=5, guessfwhm=3, guesspeak=1):
    """
    Fits a 2d airy function to the data at point (xguess, yguess)

    Args:
        frame: the data - Array of size (y,x)
        xguess,yguess: location to fit the 2d guassian to (should be pretty accurate)
        searchrad: 1/2 the length of the box used for the fit
        guessfwhm: approximate fwhm to fit to

    Returns:
        peakflux: the peakflux of the Airy function
        fwhm: diameter between the first minima along one axis
        xfit: x position
        yfit: y position
    """
    if not isinstance(searchrad, int):
        raise ValueError("searchrad needs to be an integer")

    x0 = np.rint(xguess).astype(int)
    y0 = np.rint(yguess).astype(int)
    #construct our searchbox
    fitbox = np.copy(frame[y0-searchrad:y0+searchrad+1, x0-searchrad:x0+searchrad+1])

    #mask bad pixels
    fitbox[np.where(np.isnan(fitbox))] = 0
 
    # construct an Airy function to fit to the data
    airy_psf = modeling.functional_models.AiryDisk2D()
    fity, fitx = np.indices(fitbox.shape)

    def airy_cost_function(p):
        x,y = p[0], p[1]
        if x < 0 or x > (2*searchrad+1):
            return np.inf 
        if y < 0 or y > (2*searchrad+1):
            return np.inf 
        fwhm = p[3]
        if fwhm > 2*searchrad+1:
            return np.inf

        return np.nansum(np.abs(airy_psf.evaluate(fitx, fity, p[2], p[0], p[1], p[3]/2.) + p[4] - fitbox))

    #do a least squares fit. Note that we use searchrad for x and y centers since we're narrowed it to a box of size
    #(2searchrad+1,2searchrad+1)

    guess = np.array((searchrad, searchrad, guesspeak, guessfwhm, 0))

    #p, success = optimize.leastsq(errorfunction, guess)
    res = optimize.minimize(airy_cost_function, guess, method="Nelder-Mead")

    p = res.x

    xfit = p[0]
    yfit = p[1]
    peakflux = p[2]
    fwhm = p[3] 

    # convert xfit, yfit back to image coordinates
    xfit = xfit - searchrad + x0
    yfit = yfit - searchrad + y0

    return peakflux, fwhm, xfit, yfit


def LSQ_gauss2d(planet_image, x_grid, y_grid,a,x_cen,y_cen,sig):
    """
    Calculate the squared norm of the residuals of the model with the data.
    Helper function for least square fit.
    The model is a 2d symmetric gaussian.

    Args:
        planet_image: stamp image (y,x) of the satellite spot.
        x_grid: x samples grid as given by meshgrid.
        y_grid: y samples grid as given by meshgrid.
        a: amplitude of the 2d gaussian
        x_cen: x center of the gaussian
        y_cen: y center of the gaussian
        sig: standard deviation of the gaussian

    Returns:
        Squared norm of the residuals
    """
    #gauss2d(x, y, amplitude = 1.0, xo = 0.0, yo = 0.0, sigma_x = 1.0, sigma_y = 1.0, theta = 0, offset = 0)
    model = gauss2d(x_cen,y_cen, a, sig)(x_grid, y_grid)
    #model = gauss2d(x_grid, y_grid,a,x_cen,y_cen,sig,sig,0.0)
    return np.nansum((planet_image-model)**2,axis = (0,1))#/y_model


def PSFcubefit(frame, xguess, yguess, searchrad=10,psfs_func_list=None,wave_index=None,residuals=False,rmbackground=True,add_background2residual=False):
    """
    Estimate satellite spot amplitude (peak value) by fitting a symmetric 2d gaussian.
    Fit parameters: x,y position, amplitude, standard deviation (same in x and y direction)

    Args:
        frame: the data - Array of size (y,x)
        xguess: x location to fit the 2d guassian to.
        yguess: y location to fit the 2d guassian to.
        searchrad: 1/2 the length of the box used for the fit
        psfs_func_list: List of spline fit function for the PSF_cube.
        wave_index: Index of the current wavelength. In [0,36] for GPI. Only used when psfs_func_list is not None.
        residuals: If True (Default = False) then calculate the residuals of the sat spot fit (gaussian or PSF cube).
        rmbackground: If true (default), remove any background slope to the data stamp.
        add_background2residual: If True (default is false) and if rmbackground was true, it adds the background that
                        was removed to the returned residuals.

    Returns:
        returned_flux: scalar, Estimation of the peak flux of the satellite spot.
            ie Amplitude of the fitted gaussian.
    """
    x0 = int(np.round(xguess))
    y0 = int(np.round(yguess))
    #construct our searchbox
    if (y0-searchrad)<0 or (y0+searchrad+1)>=frame.shape[0] or (x0-searchrad)<0 or (x0+searchrad+1)>=frame.shape[1]:
        # To close to the edge
        if residuals:
            return None,None
        else:
            return None
    fitbox = np.copy(frame[y0-searchrad:y0+searchrad+1, x0-searchrad:x0+searchrad+1])

    xguess_box = xguess-x0 + searchrad
    yguess_box = yguess-y0 + searchrad

    xfitbox, yfitbox = np.meshgrid(np.arange(0,2* searchrad+1, 1.0)-xguess_box, np.arange(0, 2*searchrad+1, 1.0)-yguess_box)
    stamp_r = np.sqrt((xfitbox)**2+(yfitbox)**2)
    small_aper_indices = np.where(stamp_r<3)
    big_aper_indices = np.where(stamp_r<7)

    # try to remove background
    if rmbackground:
        stamp_masked = copy(fitbox)
        stamp_x_masked = copy(xfitbox)
        stamp_y_masked = copy(yfitbox)
        stamp_masked[big_aper_indices] = np.nan
        stamp_x_masked[big_aper_indices] = np.nan
        stamp_y_masked[big_aper_indices] = np.nan
        background_med =  np.nanmedian(stamp_masked)
        stamp_masked = stamp_masked - background_med
        #Solve 2d linear fit to remove background
        xx = np.nansum(stamp_x_masked**2)
        yy = np.nansum(stamp_y_masked**2)
        xy = np.nansum(stamp_y_masked*stamp_x_masked)
        xz = np.nansum(stamp_masked*stamp_x_masked)
        yz = np.nansum(stamp_y_masked*stamp_masked)
        #Cramer's rule
        a = (xz*yy-yz*xy)/(xx*yy-xy*xy)
        b = (xx*yz-xy*xz)/(xx*yy-xy*xy)
        background =(a*(xfitbox)+b*(yfitbox) + background_med)
        fitbox = fitbox - background

    if isinstance(wave_index,(np.ndarray)):
        # Get a deprecation warning when wave_index = [5] instead of an integer. So this picks the integer...
        new_wave_index = wave_index[0]
    else:
        new_wave_index = wave_index
    model = psfs_func_list[new_wave_index](np.arange(0,2* searchrad+1, 1.0)-xguess_box,np.arange(0, 2*searchrad+1, 1.0)-yguess_box).transpose()
    # model = psfs_func_list[wave_index](np.arange(0,2* searchrad+1, 1.0)+(xguess-x0) - searchrad,np.arange(0, 2*searchrad+1, 1.0)+(yguess-y0) - searchrad)#.transpose()

    returned_flux = np.nansum(model[small_aper_indices]*fitbox[small_aper_indices])/np.nansum(model[small_aper_indices]**2)*model[searchrad,searchrad]

    if residuals:
        residuals_map = fitbox - returned_flux*model/model[searchrad,searchrad]
        if add_background2residual and rmbackground:
            residuals_map = residuals_map + background
        return returned_flux,residuals_map
    else:
        return returned_flux


def gaussfit2dLSQ(frame, xguess, yguess, searchrad=5,fit_centroid = False,residuals=False):
    """
    Estimate satellite spot amplitude (peak value) by fitting a symmetric 2d gaussian.
    Fit parameters: x,y position, amplitude, standard deviation (same in x and y direction)

    Args:
        frame: the data - Array of size (y,x)
        xguess: x location to fit the 2d guassian to.
        yguess: y location to fit the 2d guassian to.
        searchrad: 1/2 the length of the box used for the fit
        fit_centroid: If False (default), disable the centroid fit and only fit the amplitude and the standard deviation
        residuals: If True (Default = False) then calculate the residuals of the sat spot fit (gaussian or PSF cube).

    Returns:
        returned_flux: scalar, estimation of the peak flux of the satellite spot.
            ie Amplitude of the fitted gaussian.
    """
    x0 = int(np.round(xguess))
    y0 = int(np.round(yguess))
    #construct our searchbox
    fitbox = np.copy(frame[y0-searchrad:y0+searchrad+1, x0-searchrad:x0+searchrad+1])

    xguess_box = xguess-x0 + searchrad
    yguess_box = yguess-y0 + searchrad

    xfitbox, yfitbox = np.meshgrid(np.arange(0,2* searchrad+1, 1.0)-xguess_box, np.arange(0, 2*searchrad+1, 1.0)-yguess_box)

    if fit_centroid:
        param0 = [fitbox[searchrad,searchrad],0.,0.,1.5]
        LSQ_func = lambda para: LSQ_gauss2d(fitbox,xfitbox, yfitbox,para[0],para[1],para[2],para[3])

        param_fit = minimize(LSQ_func,param0, method="Nelder-Mead").x

        if (param_fit[1]**2+param_fit[2]**2)>2.**2 or (param_fit[0]/param0[0] >10.) or param_fit[3] > 3. or param_fit[3] < 0.5:
            returned_flux = np.nan
        else:
            returned_flux = param_fit[0]

        if residuals:
            model = gauss2d(param_fit[1],param_fit[2], param_fit[0], param_fit[3])(xfitbox, yfitbox)
            residuals_map = fitbox - model
            return returned_flux,residuals_map
        else:
            return returned_flux
    else:
        param0 = [fitbox[searchrad,searchrad],1.5]
        LSQ_func = lambda para: LSQ_gauss2d(fitbox,xfitbox, yfitbox,para[0],0,0,para[1])

        param_fit = minimize(LSQ_func,param0, method="Nelder-Mead").x

        # print(param_fit[1])
        if abs(param_fit[0] - fitbox[searchrad,searchrad]) > 0.5*fitbox[searchrad,searchrad]:
            returned_flux = np.nan
        else:
            returned_flux = param_fit[0]

        if residuals:
            model = gauss2d(0,0, param_fit[0], param_fit[1])(xfitbox, yfitbox)
            residuals_map = fitbox - model
            return returned_flux,residuals_map
        else:
            return returned_flux


def retrieve_planet_flux(frames, centers, astr_hdrs, sep, pa, searchrad=7, guessfwhm=3.0, guesspeak=1, refinefit=False,
                         thetas=None):
    """
    Retrives the planet flux from a series of frames given a separation and PA

    Args:
        frames: frames of data to retrieve planet. Can be a single 2-D image ([y,x]) for a series/cube ([N,y,x])
        centers: coordiantes of the image center. Can be [2]-element lst or an array that matches array of frames [N,2]
        astr_hdrs: astr_hdrs, can be a single one or an array of N of them
        sep: radial distance in pixels
        PA: parallactic angle in degrees
        searchrad: 1/2 the length of the box used for the fit
        guessfwhm: approximate fwhm to fit to
        guesspeak: approximate flux
        refinefit: whether or not to refine the positioning of the planet
        thetas: ignore PA, supply own thetas (CCW angle from +x axis toward +y)
                single number or array of size N

    Returns:
        peakflux: either a single peak flux or an array depending on whether a single frame or multiple frames
                    where passed in
    """
    measured = retrieve_planet(frames, centers, astr_hdrs, sep, pa, searchrad, guessfwhm, guesspeak, refinefit, thetas)

    if np.ndim(measured) == 1:
        # just one frame, return one number
        return measured[0]
    else:
        # return an array of fluxes
        return measured[:, 0]


def retrieve_planet(frames, centers, astr_hdrs, sep, pa, searchrad=7, guessfwhm=3.0, guesspeak=1, refinefit=True,
                    thetas=None):
    """
    Retrives the planet properties from a series of frames given a separation and PA

    Args:
        frames: frames of data to retrieve planet. Can be a single 2-D image ([y,x]) for a series/cube ([N,y,x])
        centers: coordiantes of the image center. Can be [2]-element lst or an array that matches array of frames [N,2]
        astr_hdrs: astr_hdrs, can be a single one or an array of N of them
        sep: radial distance in pixels
        PA: parallactic angle in degrees
        searchrad: 1/2 the length of the box used for the fit
        guessfwhm: approximate fwhm to fit to
        guesspeak: approximate flux
        refinefit: whether or not to refine the positioning of the planet
        thetas: ignore PA, supply own thetas (CCW angle from +x axis toward +y)
                single number or array of size N

    Returns:
        measured: (peakflux, x, y, fwhm). A single tuple if one frame passed in. Otherwise an array of tuples
    """

    # check to make sure all arguments are the right/consistent dimensions
    # get the number of dimensions of frames as reference for all the other variables
    frames_ndim = np.ndim(frames)
    if frames_ndim == 2:
        # make sure all other arguments are consistent in dimensions
        if np.ndim(centers) != 1:
            raise IndexError("centers needs to be a 2-element list")
        if np.ndim(astr_hdrs) != 0:
            raise IndexError("astr_hdrs cannot be a list because you only passed in one frame")
        if thetas is not None:
            if np.ndim(thetas) != 0:
                raise IndexError("thetas cannot be a list because you only passed in one frame")
            thetas = [thetas]
        # turn them into lists so we can reuse the same code
        frames = [frames]
        centers = [centers]
        astr_hdrs = [astr_hdrs]
    elif frames_ndim == 3:
        # make sure all other arguments are consistent in dimensions
        if np.ndim(centers) != 2:
            raise IndexError("centers needs to an array of [x,y] coordinates")
        if np.ndim(astr_hdrs) != 1:
            raise IndexError("astr_hdrs must be a list because you only passed in multiple frames")
        if thetas is not None:
            if np.ndim(thetas) != 1:
                raise IndexError("thetas must be a list because you only passed in multiple frames")
    else:
        raise IndexError("frames is either 2-D or 3-D, not {0)-D".format(frames_ndim))

    measured = []
   
    if thetas is None:
        thetas = np.array([convert_pa_to_image_polar(pa, astr_hdr) for astr_hdr in astr_hdrs])

    # loop over all of them
    for frame, center, theta in zip(frames, centers, thetas):
        # find the pixel location on this image
        # theta = covert_pa_to_image_polar(pa, astr_hdr)
        x = sep*np.cos(np.radians(theta)) + center[0]
        y = sep*np.sin(np.radians(theta)) + center[1]
        # calculate the flux
        flux, fwhm, xfit, yfit = gaussfit2d(frame, x, y, searchrad=searchrad, guessfwhm=guessfwhm, guesspeak=guesspeak, refinefit=refinefit)
        measured.append((flux, xfit, yfit, fwhm))

    # return a single number if onyl one frame passed in
    if frames_ndim == 2:
        measured = measured[0]
    else:
        measured = np.array(measured)

    return measured



__author__ = 'jruffio'
import multiprocessing as mp
import ctypes

import numpy as np
import pyklip.spectra_management as spec
import os

from pyklip.fmlib.nofm import NoFM
import pyklip.fm as fm
import pyklip.fakes as fakes

from scipy import interpolate, linalg
from copy import copy

#import matplotlib.pyplot as plt
debug = False


class ExtractSpec(NoFM):
    """
    Planet Characterization class. Goal to characterize the astrometry and photometry of a planet
    """
    def __init__(self, inputs_shape,
                 numbasis,
                 sep, pa,
                 input_psfs,
                 input_psfs_wvs,
                 datatype="float",
                 stamp_size = None):
        """
        Defining the planet to characterizae

        Args:
            inputs_shape: shape of the inputs numpy array. Typically (N, y, x)
            numbasis: 1d numpy array consisting of the number of basis vectors to use
            sep: separation of the planet
            pa: position angle of the planet
            input_psfs: the psf of the image. A numpy array with shape (wv, y, x)
            input_psfs_wvs: the wavelegnths that correspond to the input psfs
            flux_conversion: an array of length N to convert from contrast to DN for each frame. Units of DN/contrast
            wavelengths: wavelengths of data. Can just be a string like 'H' for H-band
            spectrallib: if not None, a list of spectra
            star_spt: star spectral type, if None default to some random one
            refine_fit: refine the separation and pa supplied
        """
        # allocate super class
        super(ExtractSpec, self).__init__(inputs_shape, np.array(numbasis))

        if stamp_size is None:
            self.stamp_size = 10
        else:
            self.stamp_size = stamp_size

        if datatype=="double":
            self.data_type = ctypes.c_double
        elif datatype=="float":
            self.data_type = ctypes.c_float

        self.N_numbasis =  np.size(numbasis)
        self.ny = self.inputs_shape[1]
        self.nx = self.inputs_shape[2]
        self.N_frames = self.inputs_shape[0]

        self.inputs_shape = inputs_shape
        self.numbasis = numbasis
        self.sep = sep
        self.pa = pa


        # 2018-04-10 AG: Generalizing so everything is in units of input PSF
        # Specifically removing the normlization of the input PSF
        self.input_psfs = input_psfs
        self.input_psfs_wvs = list(np.array(input_psfs_wvs,dtype=self.data_type))
        self.nl = np.size(input_psfs_wvs)

        # Make sure the peak value is unity for all wavelengths
        #self.sat_spot_spec = np.nanmax(self.input_psfs,axis=(1,2))
        #self.aper_over_peak_ratio = np.zeros(np.size(self.input_psfs_wvs))
        #for l_id in range(self.input_psfs.shape[0]):
        #    self.aper_over_peak_ratio[l_id] = np.nansum(self.input_psfs[l_id,:,:])/self.sat_spot_spec[l_id]
        #    self.input_psfs[l_id,:,:] = self.input_psfs[l_id,:,:]/np.nansum(self.input_psfs[l_id,:,:])

        self.nl, self.ny_psf, self.nx_psf =  self.input_psfs.shape

        self.psf_centx_notscaled = {}
        self.psf_centy_notscaled = {}

        numwv,ny_psf,nx_psf =  self.input_psfs.shape
        x_psf_grid, y_psf_grid = np.meshgrid(np.arange(nx_psf * 1.)-nx_psf//2,np.arange(ny_psf* 1.)-ny_psf//2)
        psfs_func_list = []
        for wv_index in range(numwv):
            model_psf = self.input_psfs[wv_index, :, :] #* self.flux_conversion * self.spectrallib[0][wv_index] * self.dflux
            psfs_func_list.append(interpolate.LSQBivariateSpline(x_psf_grid.ravel(),y_psf_grid.ravel(),model_psf.ravel(),x_psf_grid[0,0:nx_psf-1]+0.5,y_psf_grid[0:ny_psf-1,0]+0.5))

        self.psfs_func_list = psfs_func_list


    # def alloc_interm(self, max_sector_size, numsciframes):
    #     """Allocates shared memory array for intermediate step
    #
    #     Intermediate step is allocated for a sector by sector basis
    #
    #     Args:
    #         max_sector_size: number of pixels in this sector. Max because this can be variable. Stupid rotating sectors
    #
    #     Returns:
    #         interm: mp.array to store intermediate products from one sector in
    #         interm_shape:shape of interm array (used to convert to numpy arrays)
    #
    #     """
    #
    #     interm_size = max_sector_size * np.size(self.numbasis) * numsciframes * len(self.spectrallib)
    #
    #     interm = mp.Array(ctypes.c_double, interm_size)
    #     interm_shape = [numsciframes, len(self.spectrallib), max_sector_size, np.size(self.numbasis)]
    #
    #     return interm, interm_shape


    def alloc_fmout(self, output_img_shape):
        """
        Allocates shared memory for the output of the shared memory

        Args:
            output_img_shape: shape of output image (usually N,y,x,b)

        Returns:
            fmout: mp.array to store FM data in
            fmout_shape: shape of FM data array

        """

        # The 3rd dimension (self.N_frames corresponds to the spectrum)
        # The +1 in (self.N_frames+1) is for the klipped image
        fmout_size = self.N_numbasis*self.N_frames*(self.N_frames+1)*self.stamp_size*self.stamp_size
        # 2018-04-10 AG: force fmout to be type int
        fmout = mp.Array(self.data_type, int(fmout_size))
        # fmout shape is defined as:
        #   (self.N_numbasis,self.N_frames,(self.N_frames+1),self.stamp_size*self.stamp_size)
        # 1st dim: The size of the numbasis input. numasis gives the list of the number of KL modes we want to try out
        #           e.g. numbasis = [10,20,50].
        # 2nd dim: It is the Forward model dimension. It contains the forard model for each frame in the dataset.
        #           N_frames = N_cubes*(Number of spectral channel=37)
        # 3nd dim: It contains both the "spectral dimension" and the klipped image.
        #           The regular klipped data is fmout[:,:, -1,:]
        #           The regular forward model is fmout[:,:, 0:self.N_frames,:]
        #           Multiply a vector of fluxes to this dimension of fmout[:,:, 0:self.N_frames,:] and you should get
        #           forward model for that given spectrum.
        # 4th dim: pixels value. It has the size of the number of pixels in the stamp self.stamp_size*self.stamp_size.
        fmout_shape = (self.N_numbasis,self.N_frames,(self.N_frames+1),self.stamp_size*self.stamp_size )

        return fmout, fmout_shape


    # def alloc_perturbmag(self, output_img_shape, numbasis):
    #     """
    #     Allocates shared memory to store the fractional magnitude of the linear KLIP perturbation
    #     Stores a number for each frame = max(oversub + selfsub)/std(PCA(image))
    #
    #     Args:
    #         output_img_shape: shape of output image (usually N,y,x,b)
    #         numbasis: array/list of number of KL basis cutoffs requested
    #
    #     Returns:
    #         perturbmag: mp.array to store linaer perturbation magnitude
    #         perturbmag_shape: shape of linear perturbation magnitude
    #
    #     """
    #     perturbmag_shape = (output_img_shape[0], np.size(numbasis))
    #     perturbmag = mp.Array(ctypes.c_double, np.prod(perturbmag_shape))
    #
    #     return perturbmag, perturbmag_shape


    def generate_models(self, input_img_shape, section_ind, pas, wvs, radstart, radend, phistart, phiend, padding, ref_center, parang, ref_wv, flipx, stamp_size = None):
        """
        Generate model PSFs at the correct location of this segment for each image denoated by its wv and parallactic angle

        Args:
            pas: array of N parallactic angles corresponding to N images [degrees]
            wvs: array of N wavelengths of those images
            radstart: radius of start of segment
            radend: radius of end of segment
            phistart: azimuthal start of segment [radians]
            phiend: azimuthal end of segment [radians]
            padding: amount of padding on each side of sector
            ref_center: center of image
            parang: parallactic angle of input image [DEGREES]
            ref_wv: wavelength of science image
            stamp_size: size of the stamp for spectral extraction
            flipx: if True, flip x coordinate in final image

        Return:
            models: array of size (N, p) where p is the number of pixels in the segment
        """
        # create some parameters for a blank canvas to draw psfs on
        nx = input_img_shape[1]
        ny = input_img_shape[0]
        x_grid, y_grid = np.meshgrid(np.arange(nx * 1.)-ref_center[0], np.arange(ny * 1.)-ref_center[1])


        numwv, ny_psf, nx_psf =  self.input_psfs.shape

        # create bounds for PSF stamp size
        row_m = np.floor(ny_psf/2.0)    # row_minus
        row_p = np.ceil(ny_psf/2.0)     # row_plus
        col_m = np.floor(nx_psf/2.0)    # col_minus
        col_p = np.ceil(nx_psf/2.0)     # col_plus

        if stamp_size is not None:
            stamp_mask = np.zeros((ny,nx))
            # create bounds for spectral extraction stamp size
            row_m_stamp = np.floor(stamp_size/2.0)    # row_minus
            row_p_stamp = np.ceil(stamp_size/2.0)     # row_plus
            col_m_stamp = np.floor(stamp_size/2.0)    # col_minus
            col_p_stamp = np.ceil(stamp_size/2.0)     # col_plus
            stamp_indices=[]

        # a blank img array of write model PSFs into
        whiteboard = np.zeros((ny,nx))
        if debug:
            canvases = []
        models = []
        #print(self.input_psfs.shape)
        for pa, wv in zip(pas, wvs):
            #print(self.pa,self.sep)
            #print(pa,wv)
            # grab PSF given wavelength
            wv_index = spec.find_nearest(self.input_psfs_wvs,wv)[1]
            #model_psf = self.input_psfs[wv_index[0], :, :] #* self.flux_conversion * self.spectrallib[0][wv_index] * self.dflux

            # find center of psf
            # to reduce calculation of sin and cos, see if it has already been calculated before
            if pa not in self.psf_centx_notscaled:
                # flipx requires the opposite rotation
                sign = -1.
                if flipx:
                    sign = 1.
                self.psf_centx_notscaled[pa] = self.sep * np.cos(np.radians(90. - sign*self.pa - pa))
                self.psf_centy_notscaled[pa] = self.sep * np.sin(np.radians(90. - sign*self.pa - pa))
            psf_centx = (ref_wv/wv) * self.psf_centx_notscaled[pa]
            psf_centy = (ref_wv/wv) * self.psf_centy_notscaled[pa]

            # create a coordinate system for the image that is with respect to the model PSF
            # round to nearest pixel and add offset for center
            l = round(psf_centx + ref_center[0])
            k = round(psf_centy + ref_center[1])
            # recenter coordinate system about the location of the planet
            x_vec_stamp_centered = x_grid[0, int(l-col_m):int(l+col_p)]-psf_centx
            y_vec_stamp_centered = y_grid[int(k-row_m):int(k+row_p), 0]-psf_centy
            # rescale to account for the align and scaling of the refernce PSFs
            # e.g. for longer wvs, the PSF has shrunk, so we need to shrink the coordinate system
            x_vec_stamp_centered /= (ref_wv/wv)
            y_vec_stamp_centered /= (ref_wv/wv)

            # use intepolation spline to generate a model PSF and write to temp img
            whiteboard[int(k-row_m):int(k+row_p), int(l-col_m):int(l+col_p)] = \
                    self.psfs_func_list[int(wv_index)](x_vec_stamp_centered,y_vec_stamp_centered).transpose()

            # write model img to output (segment is collapsed in x/y so need to reshape)
            whiteboard.shape = [input_img_shape[0] * input_img_shape[1]]
            segment_with_model = copy(whiteboard[section_ind])
            whiteboard.shape = [input_img_shape[0],input_img_shape[1]]

            models.append(segment_with_model)
            if stamp_size is not None:
                # These are actually indices of indices. they indicate which indices correspond to the stamp in section_ind
                stamp_mask[int(k-row_m_stamp):int(k+row_p_stamp), int(l-col_m_stamp):int(l+col_p_stamp)] = 1
                stamp_mask.shape = [nx*ny]
                stamp_indices.append(np.where(stamp_mask[section_ind] == 1)[0])
                stamp_mask.shape = [ny,nx]
                stamp_mask[int(k-row_m_stamp):int(k+row_p_stamp), int(l-col_m_stamp):int(l+col_p_stamp)] = 0

        if stamp_size is not None:
            return np.array(models),stamp_indices
        else:
            return np.array(models)




    def fm_from_eigen(self, klmodes=None, evals=None, evecs=None, input_img_shape=None, input_img_num=None, ref_psfs_indicies=None, section_ind=None,section_ind_nopadding=None, aligned_imgs=None, pas=None,
                     wvs=None, radstart=None, radend=None, phistart=None, phiend=None, padding=None,IOWA = None, ref_center=None,
                     parang=None, ref_wv=None, numbasis=None, fmout=None, perturbmag=None, klipped=None, flipx=True, **kwargs):
        """
        Generate forward models using the KL modes, eigenvectors, and eigenvectors from KLIP. Calls fm.py functions to
        perform the forward modelling

        Args:
            klmodes: unpertrubed KL modes
            evals: eigenvalues of the covariance matrix that generated the KL modes in ascending order
                   (lambda_0 is the 0 index) (shape of [nummaxKL])
            evecs: corresponding eigenvectors (shape of [p, nummaxKL])
            input_image_shape: 2-D shape of inpt images ([ysize, xsize])
            input_img_num: index of sciece frame
            ref_psfs_indicies: array of indicies for each reference PSF
            section_ind: array indicies into the 2-D x-y image that correspond to this section.
                         Note needs be called as section_ind[0]
            pas: array of N parallactic angles corresponding to N reference images [degrees]
            wvs: array of N wavelengths of those referebce images
            radstart: radius of start of segment
            radend: radius of end of segment
            phistart: azimuthal start of segment [radians]
            phiend: azimuthal end of segment [radians]
            padding: amount of padding on each side of sector
            IOWA: tuple (IWA,OWA) where IWA = Inner working angle and OWA = Outer working angle both in pixels.
                It defines the separation interva in which klip will be run.
            ref_center: center of image
            numbasis: array of KL basis cutoffs
            parang: parallactic angle of input image [DEGREES]
            ref_wv: wavelength of science image
            fmout: numpy output array for FM output. Shape is (N, y, x, b)
            perturbmag: numpy output for size of linear perturbation. Shape is (N, b)
            klipped: PSF subtracted image. Shape of ( size(section), b)
            kwargs: any other variables that we don't use but are part of the input
        """
        sci = aligned_imgs[input_img_num, section_ind[0]]
        refs = aligned_imgs[ref_psfs_indicies, :]
        refs = refs[:, section_ind[0]]


        # generate models for the PSF of the science image
        model_sci, stamp_indices = self.generate_models(input_img_shape, section_ind, [parang], [ref_wv], radstart, radend, phistart, phiend, padding, ref_center, parang, ref_wv, flipx, stamp_size=self.stamp_size)
        model_sci = model_sci[0]
        stamp_indices = stamp_indices[0]

        # generate models of the PSF for each reference segments. Output is of shape (N, pix_in_segment)
        models_ref = self.generate_models(input_img_shape, section_ind, pas, wvs, radstart, radend, phistart, phiend, padding, ref_center, parang, ref_wv, flipx)

        # using original Kl modes and reference models, compute the perturbed KL modes (spectra is already in models)
        #delta_KL = fm.perturb_specIncluded(evals, evecs, klmodes, refs, models_ref)
        delta_KL_nospec = fm.pertrurb_nospec(evals, evecs, klmodes, refs, models_ref)

        # calculate postklip_psf using delta_KL
        oversubtraction, selfsubtraction = fm.calculate_fm(delta_KL_nospec, klmodes, numbasis, sci, model_sci, inputflux=None)
        # klipped_oversub.shape = (size(numbasis),Npix)
        # klipped_selfsub.shape = (size(numbasis),N_lambda or N_ref,N_pix)
        # klipped_oversub = Sum(<S|KL>KL)
        # klipped_selfsub = Sum(<N|DKL>KL) + Sum(<N|KL>DKL)


        # Note: The following could be used if we want to derotate the image but JB doesn't think we have to.
        # # write forward modelled PSF to fmout (as output)
        # # need to derotate the image in this step
        # for thisnumbasisindex in range(np.size(numbasis)):
        #         fm._save_rotated_section(input_img_shape, postklip_psf[thisnumbasisindex], section_ind,
        #                          fmout[input_img_num, :, :,thisnumbasisindex], None, parang,
        #                          radstart, radend, phistart, phiend, padding,IOWA, ref_center, flipx=True)


        # fmout shape is defined as:
        #   (self.N_numbasis,self.N_frames,(self.N_frames+1),self.stamp_size*self.stamp_size)
        # 1st dim: The size of the numbasis input. numasis gives the list of the number of KL modes we want to try out
        #           e.g. numbasis = [10,20,50].
        # 2nd dim: It is the Forward model dimension. It contains the forard model for each frame in the dataset.
        #           N_frames = N_cubes*(Number of spectral channel=37)
        # 3nd dim: It contains both the "spectral dimension" and the klipped image.
        #           The regular klipped data is fmout[:,:, -1,:]
        #           The regular forward model is fmout[:,:, 0:self.N_frames,:]
        #           Multiply a vector of fluxes to this dimension of fmout[:,:, 0:self.N_frames,:] and you should get
        #           forward model for that given spectrum.
        # 4th dim: pixels value. It has the size of the number of pixels in the stamp self.stamp_size*self.stamp_size.
        for k in range(self.N_numbasis):
            fmout[k,input_img_num, input_img_num,:] = fmout[k,input_img_num, input_img_num,:]+model_sci[stamp_indices]
        fmout[:,input_img_num, input_img_num,:] = fmout[:,input_img_num, input_img_num,:]-oversubtraction[:,stamp_indices]
        fmout[:,input_img_num, ref_psfs_indicies,:] = fmout[:,input_img_num, ref_psfs_indicies,:]-selfsubtraction[:,:,stamp_indices]
        fmout[:,input_img_num, -1,:] = klipped.T[:,stamp_indices]




    def cleanup_fmout(self, fmout):
        """
        After running KLIP-FM, we need to reshape fmout so that the numKL dimension is the first one and not the last

        Args:
            fmout: numpy array of ouput of FM

        Return:
            fmout: same but cleaned up if necessary
        """
        # Here we actually extract the spectrum


        return fmout

def invert_spect_fmodel(fmout, dataset, method = "JB", units = "natural",
                        scaling_factor=1.0):
    """
    A. Greenbaum Nov 2016
    
    Args:
        fmout: the forward model matrix which has structure:
               [numbasis, n_frames, n_frames+1, npix]
        dataset: from GPI.GPIData(filelist) -- typically set highpass=True also
        method: "JB" or "LP" to try the 2 different inversion methods (JB's or Laurent's)
        units: "natural" means the answer is scaled to the input PSF (default)
               fmout will be in these units.
               "scaled" means the output is scaled to "scaling_factor" argument
        scaling_factor: multiplies output spectrum and forward model, user set for 
                        desired calibration factor. units="scaled" must be set in 
                        args for this to work!
    Returns:
        A tuple containing the spectrum and the forward model
        (spectrum, forwardmodel)
        spectrum shape:(len(numbasis), nwav)
        
    """
    N_frames = fmout.shape[2] - 1 # The last element in this axis contains klipped image
    N_cubes = np.size(np.unique(dataset.filenums)) # 
    nl = N_frames // N_cubes
    stamp_N_pix = fmout.shape[-1]

    # Selection matrix (N_cubes, 1) shape
    spec_identity = np.identity(nl)
    selec = np.tile(spec_identity,(N_frames//nl, 1))

    # set up array for klipped image for each numbasis, n_frames x npix
    klipped = np.zeros((fmout.shape[0], fmout.shape[1], fmout.shape[3]))
    estim_spec = np.zeros((fmout.shape[0], nl))

    # The first dimension in fmout is numbasis, and there can be multiple of these,
    # Especially if you want to see how the spectrum behaves when you change parameters.
    # We'll also set aside an array to store the forward model matrix
    fm_coadd_mat = np.zeros((len(fmout), nl*stamp_N_pix, nl))
    for ii in range(len(fmout)):
        klipped[ii, ...] = fmout[ii,:, -1,:]
        # klipped_coadd will be coadded over N_cubes
        klipped_coadd = np.zeros((int(nl),int(stamp_N_pix)))
        for k in range(N_cubes):
            klipped_coadd = klipped_coadd + klipped[ii, k*nl:(k+1)*nl,:]
        print(klipped_coadd.shape)
        #klipped_coadd.shape = [int(nl),int(stamp_size),int(stamp_size)]
        # This is the 'raw' forward model, need to rearrange to solve FM*spec = klipped
        FM_noSpec = fmout[ii, :,:N_frames, :]

        # Move spectral dimension to the end (Effectively move pixel dimension to the middle)
        # [nframes, nframes, npix] -> [nframes, npix, nframes]
        FM_noSpec = np.rollaxis(FM_noSpec, 2, 1)

        # S^T . FM[npix, nframes, nframes] . S
        # essentially coadds over N_cubes via selection matrix
        # reduces to [nwav, npix, nwav]
        fm_noSpec_coadd = np.dot(selec.T,np.dot(np.rollaxis(FM_noSpec,1,0),selec))
        if method == "JB":
            #
            #JBR's matrix inversion adds up over all exposures, then inverts
            #
            #Back to a 2D array pixel array in the middle
            fm_noSpec_coadd.shape = [int(nl),stamp_N_pix, int(nl)]
            # Flatten over first 3 dims for the FM matrix to solve FM*spect = klipped
            fm_noSpec_coadd_mat = np.reshape(fm_noSpec_coadd,(int(nl*stamp_N_pix),int(nl)))
            # Invert the FM matrix
            pinv_fm_coadd_mat = np.linalg.pinv(fm_noSpec_coadd_mat)
            # solve via FM^-1 . klipped_PSF (flattened) << both are coadded over N_cubes
            estim_spec[ii,:]=np.dot(pinv_fm_coadd_mat, klipped_coadd.ravel())
            fm_coadd_mat[ii,:, :] = fm_noSpec_coadd_mat
        elif method == "LP":
            #
            #LP's matrix inversion adds over frames and one wavelength axis, then inverts
            #
            A = np.zeros((nl, nl))
            b = np.zeros(nl)
            fm = fm_noSpec_coadd.reshape(int(nl), stamp_N_pix, int(nl))
            fm_coadd_mat[ii,:, :] = \
                fm_noSpec_coadd.reshape(int(nl*stamp_N_pix), int(nl))
            fm = np.rollaxis(fm, 2,0)
            fm = np.rollaxis(fm, 2,1)
            data = klipped_coadd
            for q in range(nl):
                A[q,:] = np.dot(fm[q,:].T,fm[q,:])[q,:]
                b[q] = np.dot(fm[q,:].T,data[q])[q]
            estim_spec[ii,:] = np.dot(np.linalg.inv(A), b)
        elif method == "leastsq":
            # MF's suggestion of solving using a least sq function 
            # instead of matrix inversions
            
            #Back to a 2D array pixel array in the middle
            fm_noSpec_coadd.shape = [int(nl),stamp_N_pix,int(nl)]
            # Flatten over first 3 dims for the FM matrix to solve FM*spect = klipped
            fm_noSpec_coadd_mat = np.reshape(fm_noSpec_coadd,(int(nl*stamp_N_pix),int(nl)))
            # Saving the coadded FM
            fm_coadd_mat[ii,:, :] = fm_noSpec_coadd_mat

            # used leastsq solver
            results = linalg.lstsq(fm_noSpec_coadd_mat, klipped_coadd.ravel())
            # grab the spectrum, not using the other parts for now.
            estim_spec[ii,:], res, rank, s = results

        else:
            print("method not understood. Choose either JB, LP or leastsq.")

    if units=="scaled":
        return scaling_factor*estim_spec, fm_coadd_mat / scaling_factor
    else:
        return estim_spec, fm_coadd_mat




from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pyklip',
    version='2.3',
    description='pyKLIP: PSF Subtraction for Exoplanets and Disks',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://bitbucket.org/pyKLIP/pyklip',
    author='pyKLIP Developers',
    author_email='jwang@astro.berkeley.edu',
    license='BSD',
    include_package_data = True,
    packages=find_packages(),
    zip_safe=False,
    classifiers=[
        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Astronomy',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: BSD License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.8',
        ],
    keywords='KLIP PSF Subtraction Exoplanets Astronomy',
    install_requires=['numpy', 'scipy', 'astropy', 'matplotlib', 'emcee', 'corner']
    )



import os
import re
import subprocess
import glob
import astropy.io.fits as fits
import astropy.wcs as wcs
from astropy.modeling import models, fitting
import numpy as np
import scipy.ndimage as ndimage
from pyklip.parallelized import high_pass_filter_imgs
import scipy.stats
import multiprocessing as mp

import sys
from copy import copy

#different imports depending on if python2.7 or python3
if sys.version_info < (3,0):
    #python 2.7 behavior
    import ConfigParser
    from pyklip.instruments.Instrument import Data
    from pyklip.instruments.utils.nair import nMathar
else:
    import configparser as ConfigParser
    from pyklip.instruments.Instrument import Data
    from pyklip.instruments.utils.nair import nMathar


#from pyklip.instruments.P1640_support import P1640spots
#from pyklip.instruments.P1640_support import P1640utils
#from pyklip.instruments.P1640_support import P1640_cube_checker

from scipy.interpolate import interp1d

'''
#IMPORTANT NOTE:

Things that /must/ be in the headers of your sliced images:
- ROTOFF
- WLENGTH (wavelength band, in microns)
- GHSTPEAK or STARPEAK
- RA
- DEC


'''

class MagAOData(object):
    
    """
    A sequence of P1640 Data. Each P1640Data object has the following fields and functions 
    Args:
        filepaths: list of filepaths to files
        skipslices: a list of datacube slices to skip (supply index numbers e.g. [0,1,2,3])
        corefilepaths: a list of filepaths to core (i.e. unocculted) files, for contrast calc
        spot_directory: (None) path to the directory where the spot positions are stored. Defaults to P1640.ini val
    Attributes:
        input: Array of shape (N,y,x) for N images of shape (y,x)
        centers: Array of shape (N,2) for N centers in the format [x_cent, y_cent]
        filenums: Array of size N for the numerical index to map data to file that was passed in
        filenames: Array of size N for the actual filepath of the file that corresponds to the data
        PAs: Array of N for the parallactic angle rotation of the target (used for ADI) [in degrees]
        wvs: Array of N wavelengths of the images (used for SDI) [in microns]. For polarization data, defaults to "None"
        wcs: Array of N wcs astormetry headers for each image.
        IWA: a floating point scalar (not array). Specifies to inner working angle in pixels
        output: Array of shape (b, len(files), len(uniq_wvs), y, x) where b is the number of different KL basis cutoffs
        spot_flux: Array of N of average satellite spot flux for each frame
        contrast_scaling: Flux calibration factors (multiply by image to "calibrate" flux)
        flux_units: units of output data [DN, contrast]
        prihdrs: not used for P1640, set to None
        exthdrs: Array of N P1640 headers (these are written by the P1640 cube extraction pipeline)
    Methods:
        readdata(): reread in the data
        savedata(): save a specified data in the P1640 datacube format (in the 1st extension header)
        calibrate_output(): calibrates flux of self.output
    """

    ##########################
   ### Class Initialization ###
    ##########################
    #Some static variables to define the MagAO instrument
    centralwave = {} #in microns
    flux_zeropt = {}
    lenslet_scale = 1.0 #arcseconds per pixel (pixel scale)
    ifs_rotation = 0.0 #degrees CCW from +x axis to zenith
    observatory_latitude = 0.0
    ghstpeak_ratio = {}

    #read in MagAO configuration file and set these static variables
    package_directory = os.path.dirname(os.path.abspath(__file__))
    configfile = package_directory + "/" + "MagAO.ini"
    config = ConfigParser.ConfigParser()
    try:
        config.read(configfile)
        #get pixel scale
        lenslet_scale = float(config.get("instrument", "ifs_lenslet_scale")) #!
        #get IFS rotation
        ifs_rotation = float(config.get("instrument", "ifs_rotation"))
        bands = ['HA', 'CONT', 'z\'', 'r\'','i\'','Ys']
        for band in bands:
            centralwave[band] = float(config.get("instrument", "cen_wave_{0}".format(band)))
            flux_zeropt[band] = float(config.get("instrument", "zero_pt_flux_{0}".format(band)))
        observatory_latitude = float(config.get("observatory", "observatory_lat"))

        ghstpeak_ratio['line\''] = float(config.get("instrument", "ghost_ratio_HA"))
        ghstpeak_ratio['cont\''] = float(config.get("instrument", "ghost_ratio_CONT"))
        ghstpeak_ratio['z\''] = float(config.get("instrument", "ghost_ratio_z\'"))
        ghstpeak_ratio['i\''] = float(config.get("instrument", "ghost_ratio_i\'"))
    except ConfigParser.Error as e:
        print("Error reading MagAO configuration file: {0}".format(e.message))
        raise e
    
    #########################
   ###    Constructors     ###
    #########################
    def __init__(self, filepaths=None, highpass=False):
        """
        Initialization code for MagAOData
        """
        super(MagAOData, self).__init__()
        self._output = None
        if filepaths is None:
            #this won't get called if we run it normally so don't worry about it
            self._input = None
            self._centers = None
            self._filenums = None
            self._filenames = None
            self._PAs = None
            self._wvs = None
            self._wcs = None
            self._IWA = None
            self._OWA = None
            self.star_flux = None
            self.contrast_scaling = None
            self.prihdrs = None
            self._flipx = None
        else:
            self.readdata(filepaths, highpass=highpass)
    
    ##############################
   ### Instance Required Fields ###
    ##############################
    @property
    def input(self):
        return self._input
    @input.setter
    def input(self, newval):
        self._input = newval
    
    @property
    def centers(self):
        return self._centers
    @centers.setter
    def centers(self, newval):
        self._centers = newval

    @property
    def PAs(self):
        return self._PAs
    @PAs.setter
    def PAs(self, newval):
        self._PAs = newval
    
    @property
    def wvs(self):
        return self._wvs
    @wvs.setter
    def wvs(self, newval):
        self._wvs = newval
    
    @property
    def wcs(self):
        return self._wcs
    @wcs.setter
    def wcs(self, newval):
        self._wcs = newval

    @property
    def filenums(self):
        return self._filenums

    @property
    def IWA(self):
        return self._IWA
    @IWA.setter
    def IWA(self, newval):
        self._IWA = newval

    @property
    def OWA(self):
        return self._OWA
    @OWA.setter
    def OWA(self, newval):
        self._OWA = newval
    
    @property
    def output(self):
        return self._output
    @output.setter
    def output(self, newval):
        self._output = newval

    @property
    def flipx(self):
        return self._flipx
    @flipx.setter
    def flipx(self, newval):
        self._flipx = newval

    ###################
   ###    Methods    ###
    ###################
        
    def readdata(self, filepaths, highpass=False):
        """
        Method to open and read a list of MagAO data
        highpass: if True, run a Gaussian high pass filter (default size is sigma=imgsize/10)
                      can also be a number specifying FWHM of box in pixel units
        """
        if isinstance(filepaths, str):
            filepaths = [filepaths]
        print('reading data, num files: ',len(filepaths))
        data = []
        filenums = []
        filenames = []
        rot_angles = []
        wcs_hdrs = []
        wvs = []
        centers = []
        star_fluxes = []
        prihdrs = []

        # Create a threadpool for high pass filter
        pool = mp.Pool()

        for index, filepath in enumerate(filepaths):
            cube, center, pa, wv, astr_hdrs, prihdr, star_flux = _magao_process_file(self, filepath, highpass=highpass)
            data.append(cube)
            centers.append(center)
            star_fluxes.append(star_flux)
            rot_angles.append(pa)
            wvs.append(wv)
            wcs_hdrs.append(astr_hdrs)
            filenums.append([index])
            prihdrs.append(prihdr)
            filenames.append([filepath])

        # Close threadpool
        pool.close()
        pool.join()
            
        #FILENUMS IS 1D LIST
        data = np.array(data)
        dims = data.shape #should be 3D (#images by 451 by 451 for us, but depending on chosen size)
        filenums = np.array(filenums).reshape([dims[0]])
        filenames = np.array(filenames).reshape([dims[0]])
        rot_angles = np.array(rot_angles).reshape([dims[0]])
        wvs = np.array(wvs).reshape([dims[0]])
        wcs_hdrs = np.array(wcs_hdrs)
        dsize = dims[0]
        centers = np.zeros((dsize,2))
        for y in range(dsize):
            for x in range(2):
                centers[y][x] = (dims[1]-1)/2
        star_fluxes = np.array(star_fluxes)
 #       wcs_temp = [None]*dsize

        self._input = data
        self._centers = centers
        self._filenums = filenums
        self._filenames = filenames
        self._PAs = rot_angles
        self._wvs = wvs
        self._wcs = wcs_hdrs
        #IWA gets reset by GUI. This is the default value.
        self.IWA = 0
        # half the size of the array
        self.OWA = data.shape[1]/2
        #CHECK IWA AND OWA
        self.star_flux = star_fluxes
        self.contrast_scaling = 1./star_fluxes
        self.prihdrs = prihdrs
        self._flipx = False

    ##not currently used but will when we apply flux conversion
    def calibrate_output(self, img, spectral=False, units="contrast"):
        """
       Calibrates the flux of an output image. Can either be a broadband image or a spectral cube depending
        on if the spectral flag is set.

        Assumes the broadband flux calibration is just multiplication by a single scalar number whereas spectral
        datacubes may have a separate calibration value for each wavelength

        Args:
            img: unclaibrated image.
                 If spectral is not set, this can either be a 2-D or 3-D broadband image
                 where the last two dimensions are [y,x]
                 If specetral is True, this is a 3-D spectral cube with shape [wv,y,x]
            spectral: if True, this is a spectral datacube. Otherwise, it is a broadband image.
            units: currently only support "contrast" w.r.t central star

        Return:
            img: calibrated image of the same shape (this is the same object as the input!!!)
        """
        if units == "contrast":
            if spectral:
                # spectral cube, each slice needs it's own calibration
                numwvs = img.shape[0]
                img *= self.contrast_scaling[:numwvs, None, None]
            else:
                # broadband image
                img *= np.nanmean(self.contrast_scaling)

        return img
        
    def savedata(self, filepath, data, klipparams = None, filetype = None, zaxis = None, center=None, astr_hdr=None,
                 fakePlparams = None, more_keywords=None):
        """
        Save data in a GPI-like fashion. Aka, data and header are in the first extension header
        
        Inputs:
        filepath: path to file to output
        data: 2D or 3D data to save
        klipparams: a string of klip parameters
        filetype: filetype of the object (e.g. "KL Mode Cube", "PSF Subtracted Spectral Cube")
        zaxis: a list of values for the zaxis of the datacub (for KL mode cubes currently)
        astr_hdr: wcs astrometry header (None for NIRC2)
        center: center of the image to be saved in the header as the keywords PSFCENTX and PSFCENTY in pixels.
        The first pixel has coordinates (0,0)
        fakePlparams: fake planet params
        
        """
        hdulist = fits.HDUList()
        hdulist.append(fits.PrimaryHDU(header=self.prihdrs[0]))
        hdulist.append(fits.ImageHDU(data=data, name="Sci"))
        
        # save all the files we used in the reduction
        # we'll assume you used all the input files
        # remove duplicates from list
        #print("filenames = " + self._filenames)
        filenames = np.unique(self._filenames)
        nfiles = np.size(filenames)
        hdulist[0].header["DRPNFILE"] = nfiles
        for i, thispath in enumerate(filenames):
            thispath = thispath.replace("\\", '/')
            splited = thispath.split("/")
            fname = splited[-1]
#            matches = re.search('S20[0-9]{6}[SE][0-9]{4}', fname)
            filename = fname#matches.group(0)
            hdulist[0].header["FILE{0}".format(i)] = filename

        # write out psf subtraction parameters
        # get pyKLIP revision number
        pykliproot = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        # the universal_newline argument is just so python3 returns a string instead of bytes
        # this will probably come to bite me later
        try:
            pyklipver = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD'], cwd=pykliproot, universal_newlines=True).strip()
        except:
            pyklipver = "unknown"
        hdulist[0].header['PSFSUB'] = "pyKLIP"
        hdulist[0].header.add_history("Reduced with pyKLIP using commit {0}".format(pyklipver))
        #if self.creator is None:
        #    hdulist[0].header['CREATOR'] = "pyKLIP-{0}".format(pyklipver)
        #else:
        #    hdulist[0].header['CREATOR'] = self.creator
        #    hdulist[0].header.add_history("Reduced by {0}".self.creator)

        # store commit number for pyklip
        hdulist[0].header['pyklipv'] = pyklipver

        if klipparams is not None:
            hdulist[0].header['PSFPARAM'] = klipparams
            hdulist[0].header.add_history("pyKLIP reduction with parameters {0}".format(klipparams))

        if fakePlparams is not None:
            hdulist[0].header['FAKPLPAR'] = fakePlparams
            hdulist[0].header.add_history("pyKLIP reduction with fake planet injection parameters {0}".format(fakePlparams))

        if filetype is not None:
            hdulist[0].header['FILETYPE'] = filetype

        if zaxis is not None:
            #Writing a KL mode Cube
            if "KL Mode" in filetype:
                hdulist[0].header['CTYPE3'] = 'KLMODES'
                #write them individually
                for i, klmode in enumerate(zaxis):
                    hdulist[0].header['KLMODE{0}'.format(i)] = klmode
                    
        #removed astr_hdr stuff            

        #use the dataset center if none was passed in
        if center is None:
            center = self.centers[0]
        if center is not None:
            hdulist[0].header.update({'PSFCENTX':center[0],'PSFCENTY':center[1]})
            hdulist[0].header.update({'CRPIX1':center[0],'CRPIX2':center[1]})
            hdulist[0].header.add_history("Image recentered to {0}".format(str(center)))

        hdulist.writeto(filepath, overwrite=True)
        hdulist.close()

        
def _magao_process_file(self, filepath, highpass=False, pool = None):

    try:
        hdulist = fits.open(filepath)
        header = hdulist[0].header
        cube = hdulist[0].data
        prihdr = hdulist[0].header
        
        wvs = header['WLENGTH'] 

        datasize = cube.shape[1] #ours will be 2D
        center = [[(datasize-1)/2, (datasize-1)/2]]
        
        dims = cube.shape
        x, y = np.meshgrid(np.arange(dims[1], dtype=np.float32), np.arange(dims[0], dtype=np.float32))

        #not really parang, but north up angle, which is the parallactic angle + the angle of the rotator+90
        parang = [(float(header['ROTOFF'])+90)]

        #WCS STUFF --------------

        w = wcs.WCS(header=header, naxis=[1,2])
        
        #define empty cd matrix to put values in later
        w.wcs.cd= np.array([[0,0],[0,0]])
    
        #add WCS info to headers:
        header['CDELT1'] = 2.2222e-6 #coordinate increment, calculated from plate scale
        header['CDELT2'] = 2.2222e-6 #coordinate increment, calculated from plate scale 
        header['CRPIX1'] = (datasize-1)/2.0#x-coordinate of ref pixel (center)
        header['CRPIX2'] = (datasize-1)/2.0#y-coordinate of ref pixel
        header['CRVAL1'] = header['RA'] #Right ascension at ref point , calculated from simbad location of eps eri
        header['CRVAL2'] = header['DEC'] #declination at ref point , calculated from simbad location of eps eri
        header['CTYPE1']  = 'RA---TAN'           #/ First axis is Right Ascension                  
        header['CTYPE2']  = 'DEC--TAN'          # / Second axis is Declination                     
        header['CUNIT1']  = 'deg     '         #  / Units of data                                  
        header['CUNIT2']  = 'deg     '          # / Units of data                                  
        header['RADESYS'] = 'FK5     '           #/ R.A DEC coordinate system reference
        #move data to wcs data format:
        w.wcs.crpix = [header['CRPIX1'], header['CRPIX2']]
        w.wcs.cdelt = np.array([header['CDELT1'], header['CDELT2']])
        w.wcs.crval = [header['CRVAL1'], header['CRVAL2']]
        w.wcs.ctype = [header['CTYPE1'], header['CTYPE2']]

        #turns out WCS data can be wrong. recalculating using rotoff
        rotoff = float(header['ROTOFF'])
        #changed the minus sign in front of vert_angle to fix direction of derotation 
        vert_angle = (rotoff+90)#(360-rotoff)
        vert_angle = np.radians(vert_angle)
        pc = np.array([[-np.cos(vert_angle), np.sin(vert_angle)],[np.sin(vert_angle), np.cos(vert_angle)]])
        pixel_scale = self.lenslet_scale #.008 arcsec/pixel (hard coded, defined in MagAO.ini)
        cdmatrix = pc * pixel_scale /3600.
        w.wcs.cd[0,0] = cdmatrix[0,0]
        w.wcs.cd[0,1] = cdmatrix[0,1]
        w.wcs.cd[1,0] = cdmatrix[1,0]
        w.wcs.cd[1,1] = cdmatrix[1,1]
        astr_hdrs = w

        #-----------------------

        if wvs == 0.6564:
            ghst_psf = self.ghstpeak_ratio['line\'']
        if wvs == 0.6428:
            ghst_psf = self.ghstpeak_ratio['cont\'']
        if wvs ==  0.90969793:
            ghst_psf = self.ghstpeak_ratio['z\'']
        if wvs == 0.76691462:
            ghst_psf = self.ghstpeak_ratio['i\'']
        if 'GHSTPEAK' in header:
            star_flux = [[header['GHSTPEAK']/ghst_psf]]
        if 'STARPEAK' in header:
            star_flux = [[header['STARPEAK']]]
    finally:
        hdulist.close()

    # reshape cube to work with highpass filter
    cube = cube.reshape([1, cube.shape[0], cube.shape[1]]) #makes a 2D y by x image into a 1 by y by x cube

    # high pass and remeasure the satellite spot fluxes if necessary
    highpassed = False
    if isinstance(highpass, bool):
        if highpass:
            cube = high_pass_filter_imgs(cube, pool=pool)
            highpassed = True
    else:
        # should be a number
        if isinstance(highpass, (float, int)):
            highpass = float(highpass)
            fourier_sigma_size = (cube.shape[1] / (highpass)) / (2 * np.sqrt(2 * np.log(2)))
            cube = high_pass_filter_imgs(cube, filtersize=fourier_sigma_size, pool=pool)
            highpassed = True

    #reshape cube because wants to be 2D elsewhere
    cube = cube[0,:,:]

    return cube, center, parang, wvs, astr_hdrs, prihdr, star_flux

#comes from NIRC2 or maybe P1640, but not GPI
#def calc_starflux(cube, center):
 #   dims = cube.shape
 #   y, x = np.meshgrid(np.arange(dims[0]), np.arange(dims[1]))
 #   g_init = models.Gaussian2D(cube.max(), x_mean=center[0][0], y_mean=center[0][1], x_stddev=5, y_stddev=5, fixed={'x_mean':True,'y_mean':True,'theta':True})
 #   fit_g = fitting.LevMarLSQFitter()
 #   g = fit_g(g_init, y, x, cube)
 #   return [[g.amplitude]]
    
    
